@@ -11,13 +11,13 @@
 ** process-level counters and the deviations are calculated and
 ** visualized for the user.
 ** ==========================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        November 1996
 ** Linux-port:  June 2000
 ** Modified: 	May 2001 - Ported to kernel 2.4
 ** --------------------------------------------------------------------------
-** Copyright (C) 2000-2005 Gerlof Langeveld
+** Copyright (C) 2000-2012 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -59,7 +59,7 @@
 **        |              |             |              V            V 
 **      ______       _________     __________     ________     _________
 **     /      \     /         \   /          \   /        \   /         \
-**      /proc          /proc       accounting     process-     screen or
+**      /proc          /proc       accounting       task       screen or
 **                                    file        database        file
 **     \______/     \_________/   \__________/   \________/   \_________/
 **
@@ -71,20 +71,20 @@
 **
 **    -	photoproc()
 **	Takes a snapshot of the counters related to resource-usage of
-**	processes which are currently active. For this purpose the whole
+**	tasks which are currently active. For this purpose the whole
 **	task-list is read.
 **	This code is UNIX-flavor dependent; in case of Linux the counters
 **	are retrieved from /proc.
 **
 **    -	acctphotoproc()
 **	Takes a snapshot of the counters related to resource-usage of
-**	processes which have been finished during the last interval.
+**	tasks which have been finished during the last interval.
 **	For this purpose all new records in the accounting-file are read.
 **
 ** When all counters have been gathered, functions are called to calculate
 ** the difference between the current counter-values and the counter-values
 ** of the previous cycle. These functions operate on the system-level
-** as well as on the process-level counters. 
+** as well as on the task-level counters. 
 ** These differences are stored in a new structure(-table). 
 **
 **    -	deviatsyst()
@@ -92,13 +92,13 @@
 ** 	counters and the corresponding counters of the previous cycle.
 **
 **    -	deviatproc()
-**	Calculates the differences between the current process-level
+**	Calculates the differences between the current task-level
 ** 	counters and the corresponding counters of the previous cycle.
-**	The per-process counters of the previous cycle are stored in the
-**	process-database; this "database" is implemented as a linked list
-**	of process-info structures in memory (so no disk-accesses needed).
+**	The per-task counters of the previous cycle are stored in the
+**	task-database; this "database" is implemented as a linked list
+**	of task-info structures in memory (so no disk-accesses needed).
 **	Within this linked list hash-buckets are maintained for fast searches.
-**	The entire process-database is handled via a set of well-defined 
+**	The entire task-database is handled via a set of well-defined 
 ** 	functions from which the name starts with "pdb_..." (see the
 **	source-file procdbase.c).
 **	The processes which have been finished during the last cycle
@@ -115,6 +115,61 @@
 ** at runtime. 
 **
 ** $Log: atop.c,v $
+** Revision 1.49  2010/10/23 14:01:00  gerlof
+** Show counters for total number of running and sleep (S and D) threads.
+**
+** Revision 1.48  2010/10/23 08:18:15  gerlof
+** Catch signal SIGUSR2 to take a final sample and stop.
+** Needed for improved of suspend/hibernate.
+**
+** Revision 1.47  2010/04/23 12:20:19  gerlof
+** Modified mail-address in header.
+**
+** Revision 1.46  2010/04/23 09:57:28  gerlof
+** Version (flag -V) handled earlier after startup.
+**
+** Revision 1.45  2010/04/17 17:19:41  gerlof
+** Allow modifying the layout of the columns in the system lines.
+**
+** Revision 1.44  2010/04/16 13:00:23  gerlof
+** Automatically start another version of atop if the logfile to
+** be read has not been created by the current version.
+**
+** Revision 1.43  2010/03/04 10:51:10  gerlof
+** Support I/O-statistics on logical volumes and MD devices.
+**
+** Revision 1.42  2009/12/31 11:33:33  gerlof
+** Sanity-check to bypass kernel-bug showing 497 days of CPU-consumption.
+**
+** Revision 1.41  2009/12/17 10:51:31  gerlof
+** Allow own defined process line with key 'o' and a definition
+** in the atoprc file.
+**
+** Revision 1.40  2009/12/17 08:15:15  gerlof
+** Introduce branch-key to go to specific time in raw file.
+**
+** Revision 1.39  2009/12/10 13:34:32  gerlof
+** Cosmetical changes.
+**
+** Revision 1.38  2009/12/10 11:55:38  gerlof
+** Introduce -L flag for line length.
+**
+** Revision 1.37  2009/12/10 10:43:33  gerlof
+** Correct calculation of node name.
+**
+** Revision 1.36  2009/12/10 09:19:06  gerlof
+** Various changes related to redesign of user-interface.
+** Made by JC van Winkel.
+**
+** Revision 1.35  2009/11/27 15:11:55  gerlof
+** *** empty log message ***
+**
+** Revision 1.34  2009/11/27 15:07:25  gerlof
+** Give up root-priviliges at a earlier stage.
+**
+** Revision 1.33  2009/11/27 14:01:01  gerlof
+** Introduce system-wide configuration file /etc/atoprc
+**
 ** Revision 1.32  2008/01/07 10:16:13  gerlof
 ** Implement summaries for atopsar.
 **
@@ -128,7 +183,7 @@
 ** Introduction of variable supportflags.
 **
 ** Revision 1.28  2007/03/20 12:13:00  gerlof
-** Be sure that all pstat struct's are initialized with binary zeroes.
+** Be sure that all tstat struct's are initialized with binary zeroes.
 **
 ** Revision 1.27  2007/02/19 11:55:04  gerlof
 ** Bug-fix: flag -S was not recognized any more.
@@ -216,7 +271,7 @@
 **
 */
 
-static const char rcsid[] = "$Id: atop.c,v 1.32 2008/01/07 10:16:13 gerlof Exp $";
+static const char rcsid[] = "$Id: atop.c,v 1.49 2010/10/23 14:01:00 gerlof Exp $";
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -243,24 +298,30 @@ static const char rcsid[] = "$Id: atop.c,v 1.32 2008/01/07 10:16:13 gerlof Exp $
 #include "showgeneric.h"
 #include "parseable.h"
 
-#define	allflags  "ab:cde:fghijklmnopqrstuvwxyz1ABCDEFGHIJKLMNOP:QRSTUVWXYZ"
-#define	PROCCHUNK	50	/* process-entries for future expansion  */
+#define	allflags  "ab:cde:fghijklmnopqrstuvwxyz1ABCDEFGHIJKL:MNOP:QRSTUVWXYZ"
+#define	PROCCHUNK	100	/* process-entries for future expansion  */
 #define	MAXFL		64      /* maximum number of command-line flags  */
 
 /*
 ** declaration of global variables
 */
 struct utsname	utsname;
+int		utsnodenamelen;
 time_t 		pretime;	/* timing info				*/
 time_t 		curtime;	/* timing info				*/
 unsigned long	interval = 10;
 unsigned long 	sampcnt;
 char		screen;
-char		acctactive;	/* accounting active (boolean)		*/
+int		linelen  = 80;
+char		acctreason;	/* accounting not active (return val) 	*/
 char		rawname[RAWNAMESZ];
 char		rawreadflag;
+unsigned int	begintime, endtime;
 char		flaglist[MAXFL];
-char		deviatonly  = 1;
+char		deviatonly = 1;
+char      	usecolors  = 1;  /* boolean: colors for high occupation  */
+char		threadview = 0;	 /* boolean: show individual threads     */
+
 unsigned short	hertz;
 unsigned int	pagesize;
 int 		osrel;
@@ -268,6 +329,8 @@ int 		osvers;
 int 		ossub;
 
 int		supportflags;	/* supported features             	*/
+char		**argvp;
+
 
 struct visualize vis = {generic_samp, generic_error,
 			generic_end,  generic_usage};
@@ -278,38 +341,70 @@ struct visualize vis = {generic_samp, generic_error,
 static char		awaittrigger;	/* boolean: awaiting trigger */
 static unsigned int 	nsamples = 0xffffffff;
 static char		midnightflag;
-static unsigned int	begintime, endtime;
 
 /*
-** interpretation of defaults-file $HOME/.atop
+** interpretation of defaults-file /etc/atoprc and $HOME/.atop
 */
-void do_flags(char *);
-void do_interval(char *);
-void do_username(char *);
-void do_procname(char *);
-void do_maxcpu(char *);
-void do_maxdisk(char *);
-void do_maxintf(char *);
-void do_cpucritperc(char *);
-void do_memcritperc(char *);
-void do_swpcritperc(char *);
-void do_dskcritperc(char *);
-void do_netcritperc(char *);
-void do_swoutcritsec(char *);
-void do_almostcrit(char *);
-void do_atopsarflags(char *);
+static void		readrc(char *);
+
+void do_flags(char *, char *);
+void do_interval(char *, char *);
+void do_linelength(char *, char *);
+void do_username(char *, char *);
+void do_procname(char *, char *);
+void do_maxcpu(char *, char *);
+void do_maxdisk(char *, char *);
+void do_maxmdd(char *, char *);
+void do_maxlvm(char *, char *);
+void do_maxintf(char *, char *);
+void do_ownsysprcline(char *, char *);
+void do_ownallcpuline(char *, char *);
+void do_ownindivcpuline(char *, char *);
+void do_owncplline(char *, char *);
+void do_ownmemline(char *, char *);
+void do_ownswpline(char *, char *);
+void do_ownpagline(char *, char *);
+void do_owndskline(char *, char *);
+void do_ownnettransportline(char *, char *);
+void do_ownnetnetline(char *, char *);
+void do_ownnetinterfaceline(char *, char *);
+void do_ownprocline(char *, char *);
+void do_cpucritperc(char *, char *);
+void do_memcritperc(char *, char *);
+void do_swpcritperc(char *, char *);
+void do_dskcritperc(char *, char *);
+void do_netcritperc(char *, char *);
+void do_swoutcritsec(char *, char *);
+void do_almostcrit(char *, char *);
+void do_atopsarflags(char *, char *);
 
 static struct {
 	char	*tag;
-	void	(*func)(char *);
+	void	(*func)(char *, char *);
 } manrc[] = {
 	{	"flags",		do_flags		},
 	{	"interval",		do_interval		},
+	{	"linelen",		do_linelength		},
 	{	"username",		do_username		},
 	{	"procname",		do_procname		},
 	{	"maxlinecpu",		do_maxcpu		},
 	{	"maxlinedisk",		do_maxdisk		},
+	{	"maxlinemdd",		do_maxmdd		},
+	{	"maxlinelvm",		do_maxlvm		},
 	{	"maxlineintf",		do_maxintf		},
+	{	"ownallcpuline",	do_ownallcpuline	},
+	{	"ownonecpuline",	do_ownindivcpuline	},
+	{	"owncplline",		do_owncplline		},
+	{	"ownmemline",		do_ownmemline		},
+	{	"ownswpline",		do_ownswpline		},
+	{	"ownpagline",		do_ownpagline		},
+	{	"owndskline",		do_owndskline		},
+	{	"ownnettrline",		do_ownnettransportline	},
+	{	"ownnetnetline",	do_ownnetnetline	},
+	{	"ownnetifline",	        do_ownnetinterfaceline	},
+	{	"ownprocline",		do_ownprocline		},
+	{	"ownsysprcline",	do_ownsysprcline	},
+	{	"owndskline",	        do_owndskline		},
 	{	"cpucritperc",		do_cpucritperc		},
 	{	"memcritperc",		do_memcritperc		},
 	{	"swpcritperc",		do_swpcritperc		},
@@ -329,99 +424,38 @@ int
 main(int argc, char *argv[])
 {
 	register int	i;
-	int		c, nr, line=0;
+	int		c;
 	char		*p;
 	struct rlimit	rlim;
 
 	/*
-	** check if defaults-file $HOME/.atoprc present;
-	** if so, set defaults specified in this file
+	** since priviliged actions will be done later on, at this stage
+	** the root-priviliges are dropped by switching effective user-id
+	** to real user-id (security reasons)
 	*/
+        if (! droprootprivs() )
+	{
+		fprintf(stderr, "not possible to drop root privs\n");
+                exit(42);
+	}
+
+	/*
+	** preserve command arguments to allow restart of other version
+	*/
+	argvp = argv;
+
+	/*
+	** read defaults-files /etc/atoprc en $HOME/.atoprc (if any)
+	*/
+	readrc("/etc/atoprc");
+
 	if ( (p = getenv("HOME")) )
 	{
 		char path[1024];
 
 		snprintf(path, sizeof path, "%s/.atoprc", p);
 
-		/*
-		** check if this file is readable with the user's
-		** *real uid/gid* with syscall access()
-		*/
-		if ( access(path, R_OK) == 0)
-		{
-			FILE	*fp;
-			char	linebuf[256], tagname[16], tagvalue[16];
-
-			fp = fopen(path, "r");
-
-			while ( fgets(linebuf, sizeof linebuf, fp) )
-			{
-				line++;
-
-				nr = sscanf(linebuf, "%15s %15s",
-							tagname, tagvalue);
-
-				switch (nr)
-				{
-				   case 0:
-					continue;
-
-				   case 1:
-					if (tagname[0] == '#')
-						continue;
-
-					fprintf(stderr,
-						"~/.atoprc: syntax error line "
-						"%d (no value specified)\n",
-						line);
-
-					cleanstop(1);
-					break;		/* not reached */
-
-				   default:
-					if (tagname[0] == '#')
-						continue;
-					
-					if (tagvalue[0] != '#')
-						break;
-
-					fprintf(stderr,
-						"~/.atoprc: syntax error line "
-						"%d (no value specified)\n",
-						line);
-
-					cleanstop(1);
-				}
-
-				/*
-				** tag name and tag value found
-				** try to recognize tag name
-				*/
-				for (i=0; i < sizeof manrc/sizeof manrc[0]; i++)
-				{
-					if ( strcmp(tagname, manrc[i].tag) ==0)
-					{
-						manrc[i].func(tagvalue);
-						break;
-					}
-				}
-
-				/*
-				** tag name not recognized
-				*/
-				if (i == sizeof manrc/sizeof manrc[0])
-				{
-					fprintf(stderr,
-						"~/.atoprc: syntax error line "
-						"%d (tag name %s not valid)\n",
-						line, tagname);
-
-					cleanstop(1);
-				}
-			}
-
-			fclose(fp);
-		}
+		readrc(path);
 	}
 
 	/*
@@ -433,10 +467,9 @@ main(int argc, char *argv[])
 	else
 		p = argv[0];
 
-	if ( strcmp(p, "atopsar") == 0)
+	if ( memcmp(p, "atopsar", 7) == 0)
 		return atopsar(argc, argv);
 
-
 	/* 
 	** interpret command-line arguments & flags 
 	*/
@@ -458,6 +491,10 @@ main(int argc, char *argv[])
 				prusage(argv[0]);
 				break;
 
+			   case 'V':		/* version wanted ?           */
+				printf("%s\n", getstrvers());
+				exit(0);
+
 			   case 'w':		/* writing of raw data ?      */
 				if (optind >= argc)
 					prusage(argv[0]);
@@ -499,6 +536,13 @@ main(int argc, char *argv[])
 				vis.show_samp = parseout;
 				break;
 
+                           case 'L':		/* line length                */
+				if ( !numeric(optarg) )
+					prusage(argv[0]);
+
+				linelen = atoi(optarg);
+				break;
+
 			   default:		/* gather other flags */
 				flaglist[i++] = c;
 			}
@@ -507,7 +551,7 @@ main(int argc, char *argv[])
 		/*
 		** get optional interval-value and optional number of samples	
 		*/
-		if (optind < argc && optind < MAXFL)
+			if (optind < argc && optind < MAXFL)
 		{
 			if (!numeric(argv[optind]))
 				prusage(argv[0]);
@@ -536,6 +580,8 @@ main(int argc, char *argv[])
 	if ( (p = strchr(utsname.nodename, '.')) )
 		*p = '\0';
 
+	utsnodenamelen = strlen(utsname.nodename);
+
 	sscanf(utsname.release, "%d.%d.%d", &osrel, &osvers, &ossub);
 
 	/*
@@ -549,15 +595,26 @@ main(int argc, char *argv[])
 	*/
 	if (rawreadflag)
 	{
-		seteuid( getuid() );
-		rawread(begintime, endtime);
+		rawread();
 		cleanstop(0);
 	}
 
 	/*
 	** determine start-time for gathering current statistics
 	*/
-	curtime = getboot();
+	curtime = getboot() / hertz;
+
+	/*
+	** catch signals for proper close-down
+	*/
+	signal(SIGHUP,  cleanstop);
+	signal(SIGTERM, cleanstop);
+
+	/*
+	** regain the root-priviliges that we dropped at the beginning
+	** to do some priviliged work
+	*/
+	regainrootprivs();
 
 	/*
 	** lock ATOP in memory to get reliable samples (also when
@@ -575,31 +632,32 @@ main(int argc, char *argv[])
 	** during heavy CPU load);
 	** ignored if not running under superuser priviliges!
 	*/
-	(void) nice(-20);
-
-	/*
-	** catch signals for proper close-down
-	*/
-	signal(SIGHUP,  cleanstop);
-	signal(SIGTERM, cleanstop);
+	if ( nice(-20) == -1)
+		;
 
 	/*
 	** switch-on the process-accounting mechanism to register the
 	** (remaining) resource-usage by processes which have finished
 	*/
-	acctswon();
+	acctreason = acctswon();
 
 	/*
 	** determine properties (like speed) of all interfaces
 	*/
 	initifprop();
 
+	/*
+ 	** open socket to the IP layer to issue getsockopt() calls later on
+	*/
+	netatop_ipopen();
+	
 	/*
 	** since priviliged activities are finished now, there is no
 	** need to keep running under root-priviliges, so switch
 	** effective user-id to real user-id
 	*/
-	seteuid ( getuid() );
+        if (! droprootprivs() )
+		cleanstop(42);
 
 	/*
 	** start the engine now .....
@@ -617,9 +675,10 @@ main(int argc, char *argv[])
 static void
 engine(void)
 {
+	int 			i, j;
 	struct sigaction 	sigact;
 	static time_t		timelimit;
-	void			getusr1(int);
+	void			getusr1(int), getusr2(int);
 
 	/*
 	** reserve space for system-level statistics
@@ -630,15 +689,19 @@ engine(void)
 	static struct sstat	*hlpsstat;
 
 	/*
-	** reserve space for process-level statistics
+	** reserve space for task-level statistics
 	*/
-	static struct pstat	*curpact;	/* current active list  */
+	static struct tstat	*curpact;	/* current active list  */
 	static int		curplen;	/* current active size  */
 
-	struct pstat		*curpexit;	/* exitted process list	*/
-	struct pstat		*devpstat;	/* deviation list	*/
+	struct tstat		*curpexit;	/* exited process list	*/
+	struct tstat		*devtstat;	/* deviation list	*/
+	struct tstat		**devpstat;	/* pointers to processes*/
+						/* in deviation list    */
 
-	int			npresent, nexit, n, nzombie;
+	unsigned int		ntask, nexit, nexitnet;
+	unsigned int		noverflow, ndeviat, nactproc;
+	int			totproc, totrun, totslpi, totslpu, totzombie;
 
 	/*
 	** initialization: allocate required memory dynamically
@@ -647,23 +710,26 @@ engine(void)
 	presstat = calloc(1, sizeof(struct sstat));
 	devsstat = calloc(1, sizeof(struct sstat));
 
-	curplen = countprocs() + PROCCHUNK;
-	curpact = calloc(curplen, sizeof(struct pstat));
+	curplen  = countprocs() * 3 / 2;	/* add 50% for threads */
+	curpact  = calloc(curplen, sizeof(struct tstat));
 
-	if (!cursstat || !presstat || !devsstat || !curpact)
-	{
-		fprintf(stderr, "unexpected calloc-failure...\n");
-		cleanstop(1);
-	}
+	ptrverify(cursstat, "Malloc failed for current sysstats\n");
+	ptrverify(presstat, "Malloc failed for prev    sysstats\n");
+	ptrverify(devsstat, "Malloc failed for deviate sysstats\n");
+	ptrverify(curpact,  "Malloc failed for %d procstats\n", curplen);
 
 	/*
-	** install the signal-handler for ALARM and SIGUSR1 (both triggers
+	** install the signal-handler for ALARM, USR1 and USR2 (triggers
 	* for the next sample)
 	*/
 	memset(&sigact, 0, sizeof sigact);
 	sigact.sa_handler = getusr1;
 	sigaction(SIGUSR1, &sigact, (struct sigaction *)0);
 
+	memset(&sigact, 0, sizeof sigact);
+	sigact.sa_handler = getusr2;
+	sigaction(SIGUSR2, &sigact, (struct sigaction *)0);
+
 	memset(&sigact, 0, sizeof sigact);
 	sigact.sa_handler = getalarm;
 	sigaction(SIGALRM, &sigact, (struct sigaction *)0);
@@ -711,7 +777,7 @@ engine(void)
 
 		/*
 		** wait for alarm-signal to arrive (except first sample)
-		** or wait for SIGUSR1 in case of an interval of 0.
+		** or wait for SIGUSR1/SIGUSR2
 		*/
 		if (sampcnt > 0 && awaittrigger)
 			pause();
@@ -738,50 +804,111 @@ engine(void)
 		deviatsyst(cursstat, presstat, devsstat);
 
 		/*
-		** take a snapshot of the current process-level statistics 
+		** take a snapshot of the current task-level statistics 
 		** and calculate the deviations (i.e. calculate the activity
 		** during the last sample)
 		**
-		** first register active processes
+		** first register active tasks
 		**  --> atop malloc's a minimal amount of space which is
 		**      only extended when needed
 		*/
-		memset(curpact, 0, curplen * sizeof(struct pstat));
+		memset(curpact, 0, curplen * sizeof(struct tstat));
 
-		while ( (npresent = photoproc(curpact, curplen)) == curplen)
+		while ( (ntask = photoproc(curpact, curplen)) == curplen)
 		{
-			curplen = countprocs() + PROCCHUNK;
+			curplen += PROCCHUNK;
 
 			curpact = realloc(curpact,
-			                   curplen * sizeof(struct pstat));
+					curplen * sizeof(struct tstat));
 
-			memset(curpact, 0, curplen * sizeof(struct pstat));
+			ptrverify(curpact,
+			          "Realloc failed for %d tasks\n", curplen);
+
+			memset(curpact, 0, curplen * sizeof(struct tstat));
 		}
 
 		/*
-		** register processes which exited during last sample;
-		** first determine how many processes exited and
-		** reserve space for them, and secondly obtain the info
+		** register processes that exited during last sample;
+		** first determine how many processes exited
+		**
+		** the number of exited processes is limited to avoid
+		** that atop explodes in memory and introduces OOM killing
 		*/
 		nexit = acctprocnt();	/* number of exited processes */
 
-		if (nexit > 0)	
+		if (nexit > MAXACCTPROCS)
 		{
-			curpexit = malloc(  nexit * sizeof(struct pstat));
-			memset(curpexit, 0, nexit * sizeof(struct pstat));
-
-			acctphotoproc(curpexit, nexit);
+			noverflow = nexit - MAXACCTPROCS;
+			nexit     = MAXACCTPROCS;
 		}
 		else
-			curpexit = NULL;
+			noverflow = 0;
+
+		/*
+		** determine how many processes have been exited
+		** for the netatop module (only processes that have
+		** used the network)
+		*/
+		if (nexit > 0 && (supportflags & NETATOPD))
+			nexitnet = netatop_exitstore();
+		else
+			nexitnet = 0;
+
+		/*
+		** reserve space for the exited processes and read them
+		*/
+		if (nexit > 0)
+		{
+			curpexit = malloc(nexit * sizeof(struct tstat));
+
+			ptrverify(curpexit,
+			          "Malloc failed for %d exited processes\n",
+			          nexit);
+
+			memset(curpexit, 0, nexit * sizeof(struct tstat));
+
+			acctphotoproc(curpexit, nexit);
+
+			/*
+ 			** reposition offset in accounting file when not
+			** all exited processes have been read (i.e. skip
+			** those processes)
+			*/
+			if (noverflow)
+				acctrepos(noverflow);
+		}
+		else
+		{
+			curpexit    = NULL;
+		}
 
 		/*
 		** calculate deviations
 		*/
-		devpstat = malloc((npresent+nexit) * sizeof(struct pstat));
+		devtstat = malloc((ntask+nexit) * sizeof(struct tstat));
 
-		n = deviatproc(curpact, npresent, curpexit, nexit,
-					deviatonly, devpstat, &nzombie);
+		ptrverify(devtstat, "Malloc failed for %d modified processes\n",
+			          				ntask+nexit);
+
+		ndeviat = deviatproc(curpact, ntask, curpexit, nexit, 
+				deviatonly, devtstat, devsstat, &nactproc,
+				&totproc, &totrun, &totslpi, &totslpu, 
+		                &totzombie);
+
+  	      	/*
+ 		** create list of pointers specifically to the process entries
+		** in the task list
+		*/
+       		devpstat = malloc(sizeof (struct tstat *) * nactproc);
+
+		ptrverify(devpstat, "Malloc failed for %d process ptrs\n",
+			          				nactproc);
+
+		for (i=0, j=0; i < ndeviat; i++)
+		{
+			if ( (devtstat+i)->gen.isproc)
+				devpstat[j++] = devtstat+i;
+		}
 
 		/*
 		** activate the installed print-function to visualize
@@ -789,8 +916,10 @@ engine(void)
 		*/
 		lastcmd = (vis.show_samp)( curtime,
 				     curtime-pretime > 0 ? curtime-pretime : 1,
-		           	     devsstat, devpstat, n, npresent,
-		           	     nzombie, nexit, sampcnt==0);
+		           	     devsstat, devtstat, devpstat,
+		                     ndeviat, ntask, nactproc,
+		                     totproc, totrun, totslpi, totslpu,
+		                     totzombie, nexit, noverflow, sampcnt==0);
 
 		/*
 		** release dynamically allocated memory
@@ -798,19 +927,23 @@ engine(void)
 		if (nexit > 0)
 			free(curpexit);
 
+		if (nexitnet > 0)
+			netatop_exiterase();
+
+		free(devtstat);
 		free(devpstat);
 
 		if (lastcmd == 'r')	/* reset requested ? */
 		{
 			sampcnt = -1;
 
-			curtime = getboot();	/* reset current time */
+			curtime = getboot() / hertz;	// reset current time
 
 			/* set current (will be 'previous') counters to 0 */
 			memset(cursstat, 0,           sizeof(struct sstat));
-			memset(curpact,  0, curplen * sizeof(struct pstat));
+			memset(curpact,  0, curplen * sizeof(struct tstat));
 
-			/* remove all processes in process database */
+			/* remove all tasks in database */
 			pdb_makeresidue();
 			pdb_cleanresidue();
 		}
@@ -835,6 +968,8 @@ prusage(char *myname)
 	printf("\t  -%c  show or log all processes (i.s.o. active processes "
 	                "only)\n", MALLPROC);
 	printf("\t  -P  generate parseable output for specified label(s)\n");
+	printf("\t  -L  alternate line length (default 80) in case of "
+			"non-screen output\n");
 
 	(*vis.show_usage)();
 
@@ -842,6 +977,7 @@ prusage(char *myname)
 	printf("\tspecific flags for raw logfiles:\n");
 	printf("\t  -w  write raw data to   file (compressed)\n");
 	printf("\t  -r  read  raw data from file (compressed)\n");
+	printf("\t      special file: y[y...] for yesterday (repeated)\n");
 	printf("\t  -S  finish atop automatically before midnight "
 	                "(i.s.o. #samples)\n");
 	printf("\t  -b  begin showing data from specified time\n");
@@ -855,7 +991,6 @@ prusage(char *myname)
 			" (kill -USR1 pid_atop)\n");
 	printf("or with the keystroke '%c' in interactive mode.\n", MSAMPNEXT);
 
-
 	cleanstop(1);
 }
 
@@ -881,18 +1016,125 @@ getusr1(int sig)
 }
 
 /*
-** functions to handle a particular tag in the .atoprc file
+** handler for USR2-signal
 */
 void
-do_interval(char *val)
+getusr2(int sig)
 {
-	if (numeric(val))
+	awaittrigger=0;
+	nsamples = sampcnt;	// force stop after next sample
+}
+
+/*
+** functions to handle a particular tag in the .atoprc file
+*/
+extern int get_posval(char *name, char *val);
+
+void
+do_interval(char *name, char *val)
+{
+	interval = get_posval(name, val);
+}
+
+void
+do_linelength(char *name, char *val)
+{
+	linelen = get_posval(name, val);
+}
+
+/*
+** read RC-file and modify defaults accordingly
+*/
+static void
+readrc(char *path)
+{
+	int	i, nr, line=0, errorcnt = 0;
+
+	/*
+	** check if this file is readable with the user's
+	** *real uid/gid* with syscall access()
+	*/
+	if ( access(path, R_OK) == 0)
 	{
-		interval = atoi(val);
-	}
-	else
-	{
-		fprintf(stderr, ".atoprc: interval value not numeric\n");
-		exit(1);
+		FILE	*fp;
+		char	linebuf[256], tagname[20], tagvalue[256];
+
+		fp = fopen(path, "r");
+
+		while ( fgets(linebuf, sizeof linebuf, fp) )
+		{
+			line++;
+
+			i = strlen(linebuf);
+
+			if (i > 0 && linebuf[i-1] == '\n')
+				linebuf[i-1] = 0;
+
+			nr = sscanf(linebuf, "%19s %255[^#]",
+						tagname, tagvalue);
+
+			switch (nr)
+			{
+			   case 0:
+				continue;
+
+			   case 1:
+				if (tagname[0] == '#')
+					continue;
+
+				fprintf(stderr,
+					"%s: syntax error line "
+					"%d (no value specified)\n",
+					path, line);
+
+				cleanstop(1);
+				break;		/* not reached */
+
+			   default:
+				if (tagname[0] == '#')
+					continue;
+				
+				if (tagvalue[0] != '#')
+					break;
+
+				fprintf(stderr,
+					"%s: syntax error line "
+					"%d (no value specified)\n",
+					path, line);
+
+				cleanstop(1);
+			}
+
+			/*
+			** tag name and tag value found
+			** try to recognize tag name
+			*/
+			for (i=0; i < sizeof manrc/sizeof manrc[0]; i++)
+			{
+				if ( strcmp(tagname, manrc[i].tag) ==0)
+				{
+					manrc[i].func(tagname, tagvalue);
+					break;
+				}
+			}
+
+			/*
+			** tag name not recognized
+			*/
+			if (i == sizeof manrc/sizeof manrc[0])
+			{
+				fprintf(stderr,
+					"%s: warning at line %2d "
+					"- tag name %s not valid\n",
+					path, line, tagname);
+
+				errorcnt++;
+			}
+		}
+
+		if (errorcnt)
+			sleep(2);
+
+		fclose(fp);
 	}
 }
