@@ -7,11 +7,11 @@
 ** This source-file contains the 'atopsar'-functionality, that makes use
 ** of the 'atop'-framework.
 ** ==========================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        July 2007
 ** --------------------------------------------------------------------------
-** Copyright (C) 2007, 2008 Gerlof Langeveld
+** Copyright (C) 2007-2010 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -27,10 +27,9 @@
 ** along with this program; if not, write to the Free Software
 ** Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 ** --------------------------------------------------------------------------
-**
 */
 
-static const char rcsid[] = "$Id: atopsar.c,v 1.17 2008/02/26 13:44:53 gerlof Exp $";
+static const char rcsid[] = "$Id: atopsar.c,v 1.28 2010/11/26 06:19:43 gerlof Exp $";
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -70,12 +69,11 @@ static const char rcsid[] = "$Id: atopsar.c,v 1.17 2008/02/26 13:44:53 gerlof Ex
 ** miscellaneous values
 */
 static unsigned int 	nsamples = 9999999;
-static unsigned int	begintime, endtime;
 static char		stampalways;
-static char		usecolors = 't';
 static char		usemarkers;
 static char		allresources;
 static int		numreports;
+static time_t		saved_begintime;
 static unsigned int	repeathead = 9999999;
 static unsigned int	summarycnt = 1;
 static char		*datemsg = "-------------------------- analysis "
@@ -106,8 +104,10 @@ static char 		coloron;       /* boolean: colors active now      */
 static void	engine(void);
 static void	pratopsaruse(char *);
 static void	reportlive(time_t, int, struct sstat *);
-static char	reportraw (time_t, int, struct sstat *,
-			struct pstat *, int, int, int, int, char);
+static char	reportraw (time_t, int,
+		           struct sstat *, struct tstat *, struct tstat **,
+		           int, int, int, int, int, int, int, int,
+		           int, unsigned int, char);
 static void	reportheader(struct utsname *, time_t);
 static time_t	daylimit(time_t);
 
@@ -118,6 +118,8 @@ atopsar(int argc, char *argv[])
 	struct rlimit	rlim;
 	char		*p, *flaglist;
 
+	usecolors = 't';
+
 	/* 
 	** interpret command-line arguments & flags 
 	*/
@@ -128,6 +130,8 @@ atopsar(int argc, char *argv[])
 		*/
 		flaglist = malloc(pricnt+32);
 
+		ptrverify(flaglist, "Malloc failed for %d flags\n", pricnt+32);
+
 		for (i=0; i < pricnt; i++)
 			flaglist[i] = pridef[i].flag;
 
@@ -149,6 +153,8 @@ atopsar(int argc, char *argv[])
                            case 'b':		/* begin time ?          */
 				if ( !hhmm2secs(optarg, &begintime) )
 					pratopsaruse(argv[0]);
+
+				saved_begintime = begintime;
 				break;
 
                            case 'e':		/* end   time ?          */
@@ -292,8 +298,6 @@ atopsar(int argc, char *argv[])
 	*/
 	if (rawreadflag)
 	{
-		seteuid( getuid() );	/* get rid of root privileges */
-
 		/*
 		** select own reportraw-function to be called
 		** by the rawread function
@@ -304,9 +308,10 @@ atopsar(int argc, char *argv[])
 		{
 			if ( pridef[i].wanted )
 			{
-				prinow = i;
-				daylim = 0;
-				rawread(begintime, endtime);
+				prinow    = i;
+				daylim    = 0;
+				begintime = saved_begintime;
+				rawread();
 				printf("\n");
 			}
 		}
@@ -338,6 +343,12 @@ atopsar(int argc, char *argv[])
 	*/
 	curtime = time(0);
 
+        /*
+	** regain the root-priviliges that we dropped at the beginning
+	** to do some priviliged work now
+	*/
+        regainrootprivs();
+
 	/*
 	** lock in memory to get reliable samples (also when
 	** memory is low);
@@ -353,8 +364,8 @@ atopsar(int argc, char *argv[])
 	** during heavy CPU load);
 	** ignored if not running under superuser privileges!
 	*/
-	(void) nice(-20);
-
+	if ( nice(-20) == -1)
+		;
 	/*
 	** determine properties (like speed) of all interfaces
 	*/
@@ -365,7 +376,8 @@ atopsar(int argc, char *argv[])
 	** need to keep running under root-privileges, so switch
 	** effective user-id to real user-id
 	*/
-	seteuid ( getuid() );
+	if (! droprootprivs() )
+		cleanstop(42);
 
 	/*
 	** start live reporting
@@ -399,6 +411,10 @@ engine(void)
 	presstat = calloc(1, sizeof(struct sstat));
 	devsstat = calloc(1, sizeof(struct sstat));
 
+	ptrverify(cursstat,  "Malloc failed for current sysstats\n");
+	ptrverify(presstat,  "Malloc failed for prev    sysstats\n");
+	ptrverify(devsstat,  "Malloc failed for deviate sysstats\n");
+
 	/*
 	** install the signal-handler for ALARM and SIGUSR1 (both triggers
 	** for the next sample)
@@ -516,11 +532,11 @@ reportlive(time_t curtime, int numsecs, struct sstat *ss)
 			*/
 			printf("%s  ", convtime(curtime, timebuf));
 	
-			if ( !(pridef[i].priline)(ss, (struct pstat *)0, 0,
+			if ( !(pridef[i].priline)(ss, (struct tstat *)0, 0,
 				numsecs, numsecs*hertz, hertz,
 				osvers, osrel, ossub,
 				stampalways ? timebuf : "        ",
-	                        0, 0, 0) )
+	                        0, 0, 0, 0, 0, 0) )
 			{
 				/*
 				** print line has failed;
@@ -585,11 +601,11 @@ reportlive(time_t curtime, int numsecs, struct sstat *ss)
 		*/
 		printf("%s  ", convtime(curtime, timebuf));
 	
-		if ( !(rv = (pridef[i].priline)(ss, (struct pstat *)0, 0,
+		if ( !(rv = (pridef[i].priline)(ss, (struct tstat *)0, 0,
 					numsecs, numsecs*hertz, hertz,
 					osvers, osrel, ossub, 
 		                        stampalways ? timebuf : "        ",
-	                                0, 0, 0) ) )
+	                        	0, 0, 0, 0, 0, 0) ) )
 		{
 			/*
 			** print line has failed;
@@ -629,16 +645,17 @@ reportlive(time_t curtime, int numsecs, struct sstat *ss)
 */
 static char
 reportraw(time_t curtime, int numsecs,
-         	struct sstat *ss, struct pstat *ps,
-         	int nlist, int npresent, int nzombie,
-         	int nexit, char flags)
+         	struct sstat *ss, struct tstat *ts, struct tstat **proclist,
+         	int ndeviat, int ntask, int nactproc,
+		int totproc, int totrun, int totslpi, int totslpu, int totzomb,
+		int nexit, unsigned int noverflow, char flags)
 {
 	static char		firstcall = 1;
 	char			timebuf[16], datebuf[16];
 	unsigned int		rv;
 	static unsigned int	curline, headline, sampsum,
-				totalsec, totalexit,
-				lastnpres, lastnzomb;
+				totalsec, totalexit, lastnpres,
+				lastntrun, lastntslpi, lastntslpu, lastnzomb;
 	static time_t		lasttime;
 	static struct sstat	totsyst;
 
@@ -730,11 +747,12 @@ reportraw(time_t curtime, int numsecs,
 			printf("%s  ", convtime(lasttime, timebuf));
 
 			rv = (pridef[prinow].priline)(&totsyst,
-				(struct pstat *)0, 0,
+				(struct tstat *)0, 0,
 				totalsec, totalsec*hertz, hertz,
 			        osvers, osrel, ossub,
 		                stampalways ? timebuf : "        ",
-	                        lastnpres, totalexit, lastnzomb);
+	                        lastnpres, lastntrun, lastntslpi, lastntslpu,
+				totalexit, lastnzomb);
 
 			if (rv == 0)
 			{
@@ -780,11 +798,12 @@ reportraw(time_t curtime, int numsecs,
 	{
 		printf("%s  ", convtime(curtime, timebuf));
 
-		rv = (pridef[prinow].priline) (ss, ps, npresent+nexit,
+		rv = (pridef[prinow].priline) (ss, ts, proclist, nactproc,
 				numsecs, numsecs*hertz, hertz,
 				osvers, osrel, ossub,
 	               		stampalways ? timebuf : "        ",
-				npresent, nexit, nzombie);
+				ndeviat, totrun, totslpi, totslpu,
+				nexit, totzomb);
 
 		if (rv == 0)
 		{
@@ -820,8 +839,11 @@ reportraw(time_t curtime, int numsecs,
 		** contains the log-restart indicator
 		*/
 		lasttime   = curtime;
-		lastnpres  = npresent;
-		lastnzomb  = nzombie;
+		lastnpres  = totproc;
+		lastntrun  = totrun;
+		lastntslpi = totslpi;
+		lastntslpu = totslpu;
+		lastnzomb  = totzomb;
 
 		/*
 		** print line only if needed
@@ -834,11 +856,12 @@ reportraw(time_t curtime, int numsecs,
 			printf("%s  ", convtime(curtime, timebuf));
 
 			rv = (pridef[prinow].priline) (&totsyst,
-					(struct pstat *)0, 0,
+					(struct tstat *)0, 0,
 					totalsec, totalsec*hertz, hertz,
 					osvers, osrel, ossub,
 					stampalways ? timebuf : "        ",
-					npresent, totalexit, nzombie);
+					ndeviat, totrun, totslpi, totslpu,
+					totalexit, totzomb);
 
 			if (rv == 0)
 			{
@@ -920,7 +943,7 @@ pratopsaruse(char *myname)
 	fprintf(stderr,
 		"\t  -r  read statistical data from specific atop logfile\n");
 	fprintf(stderr,
-		"\t      (pathname, or date in format YYYYMMDD)\n");
+		"\t      (pathname, or date in format YYYYMMDD, or y[y..])\n");
 	fprintf(stderr,
 		"\t  -R  summarize <cnt> samples into one sample\n");
 	fprintf(stderr,
@@ -1106,15 +1129,15 @@ do_atopsarflags(char *val)
 static void
 cpuhead(int osvers, int osrel, int ossub)
 {
-	printf("cpu   %%usr  %%nice   %%sys  %%irq %%softirq   %%steal "
-	       " %%wait  %%idle  _cpu_");
+	printf("cpu  %%usr %%nice %%sys %%irq %%softirq  %%steal %%guest "
+	       " %%wait %%idle  _cpu_");
 }
 
 static int
-cpuline(struct sstat *ss, struct pstat *ps, int nproc,
+cpuline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	register int	i, nlines = 1;
 	count_t		cputot;
@@ -1126,7 +1149,8 @@ cpuline(struct sstat *ss, struct pstat *ps, int nproc,
         cputot = ss->cpu.all.stime + ss->cpu.all.utime +
                  ss->cpu.all.ntime + ss->cpu.all.itime +
                  ss->cpu.all.wtime + ss->cpu.all.Itime +
-                 ss->cpu.all.Stime + ss->cpu.all.steal;
+                 ss->cpu.all.Stime + ss->cpu.all.steal +
+                 ss->cpu.all.guest;
 
 	if (cputot == 0)
 		cputot = 1;	/* avoid divide-by-zero */
@@ -1139,13 +1163,14 @@ cpuline(struct sstat *ss, struct pstat *ps, int nproc,
 
 	preprint(badness);
 
-	printf("all %6.0lf %6.0lf %6.0lf %5.0lf %8.0lf %8.0f %6.0lf %6.0lf",
+	printf("all %5.0lf %5.0lf %4.0lf %4.0lf %8.0lf %7.0f %6.0f %6.0lf %5.0lf",
                 (double) (ss->cpu.all.utime * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.ntime * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.stime * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.Itime * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.Stime * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.steal * 100.0) / cputot * ss->cpu.nrcpu,
+                (double) (ss->cpu.all.guest * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.wtime * 100.0) / cputot * ss->cpu.nrcpu,
                 (double) (ss->cpu.all.itime * 100.0) / cputot * ss->cpu.nrcpu);
 
@@ -1161,7 +1186,8 @@ cpuline(struct sstat *ss, struct pstat *ps, int nproc,
         		cputot = ss->cpu.cpu[i].stime + ss->cpu.cpu[i].utime +
                  	         ss->cpu.cpu[i].ntime + ss->cpu.cpu[i].itime +
                  	         ss->cpu.cpu[i].wtime + ss->cpu.cpu[i].Itime +
-                 	         ss->cpu.cpu[i].Stime + ss->cpu.cpu[i].steal;
+                 	         ss->cpu.cpu[i].Stime + ss->cpu.cpu[i].steal +
+                 	         ss->cpu.cpu[i].guest;
 
 			if (cputot == 0)
 				cputot = 1;	/* avoid divide-by-zero */
@@ -1177,8 +1203,8 @@ cpuline(struct sstat *ss, struct pstat *ps, int nproc,
 
 			preprint(badness);
 
-			printf("%4d %6.0lf %6.0lf %6.0lf %5.0lf %8.0lf "
-			       "%8.0f %6.0lf %6.0lf",
+			printf("%4d %5.0lf %5.0lf %4.0lf %4.0lf %8.0lf "
+			       "%7.0f %6.0lf %6.0lf %5.0lf",
 			     i,
                  	     (double)(ss->cpu.cpu[i].utime * 100.0) / cputot,
                 	     (double)(ss->cpu.cpu[i].ntime * 100.0) / cputot,
@@ -1186,6 +1212,7 @@ cpuline(struct sstat *ss, struct pstat *ps, int nproc,
                 	     (double)(ss->cpu.cpu[i].Itime * 100.0) / cputot,
                 	     (double)(ss->cpu.cpu[i].Stime * 100.0) / cputot,
                 	     (double)(ss->cpu.cpu[i].steal * 100.0) / cputot,
+                	     (double)(ss->cpu.cpu[i].guest * 100.0) / cputot,
                 	     (double)(ss->cpu.cpu[i].wtime * 100.0) / cputot,
                 	     (double)(ss->cpu.cpu[i].itime * 100.0) / cputot);
 
@@ -1209,10 +1236,10 @@ prochead(int osvers, int osrel, int ossub)
 }
 
 static int
-procline(struct sstat *ss, struct pstat *ps, int nproc,
+procline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%7.0lf %9.0lf %9.2lf  %8.2lf %8.2lf  %8.2lf\n",
 		(double)ss->cpu.csw    / deltasec,
@@ -1233,10 +1260,10 @@ taskhead(int osvers, int osrel, int ossub)
 }
 
 static int
-taskline(struct sstat *ss, struct pstat *ps, int nproc,
+taskline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	if (ppres == 0)
 	{
@@ -1244,25 +1271,12 @@ taskline(struct sstat *ss, struct pstat *ps, int nproc,
 		return 0;
 	}
 
-	if (ps)		/* process statistics available */
+	if (ts)		/* process statistics available */
 	{
-		int		i;
-		long int	trun, tslpi, tslpu;
-
-		/*
-		** accumulate number of threads per state
-		*/
-		for (i=0, trun=tslpi=tslpu=0; i < nproc; i++)
-		{
-			trun	+= (ps+i)->gen.nthrrun;
-			tslpi	+= (ps+i)->gen.nthrslpi;
-			tslpu	+= (ps+i)->gen.nthrslpu;
-		}
-
-		printf("%8.2lf %7.2lf  %7d %7d    %6ld %7ld %7ld\n",
+		printf("%8.2lf %7.2lf  %7d %7d    %6d %7d %7d\n",
 			(double)ss->cpu.nprocs / deltasec,
 			(double)pexit          / deltasec,
-			ppres, pzombie, trun, tslpi, tslpu);
+			ppres, pzombie, ntrun, ntslpi, ntslpu);
 	}
 	else
 	{
@@ -1281,15 +1295,15 @@ taskline(struct sstat *ss, struct pstat *ps, int nproc,
 static void
 memhead(int osvers, int osrel, int ossub)
 {
-	printf("memtotal memfree buffers   cached  slabmem"
-	       "    swptotal swpfree  _mem_"             );
+	printf("memtotal memfree buffers cached dirty slabmem"
+	       "  swptotal swpfree _mem_"             );
 }
 
 static int
-memline(struct sstat *ss, struct pstat *ps, int nproc,
+memline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	unsigned int	mbadness, sbadness;
 
@@ -1310,11 +1324,12 @@ memline(struct sstat *ss, struct pstat *ps, int nproc,
 
 	preprint(mbadness >= sbadness ? mbadness : sbadness);
 
-	printf("%7lldM %6lldM %6lldM %7lldM %7lldM    %7lldM %6lldM",
+	printf("%7lldM %6lldM %6lldM %5lldM %4lldM %6lldM  %7lldM %6lldM",
 		ss->mem.physmem   * (pagesize / 1024) /1024,
 		ss->mem.freemem   * (pagesize / 1024) /1024,
 		ss->mem.buffermem * (pagesize / 1024) /1024,
 		ss->mem.cachemem  * (pagesize / 1024) /1024,
+		ss->mem.cachedrt  * (pagesize / 1024) /1024,
 		ss->mem.slabmem   * (pagesize / 1024) /1024,
 		ss->mem.totswap   * (pagesize / 1024) /1024,
 		ss->mem.freeswap  * (pagesize / 1024) /1024);
@@ -1335,10 +1350,10 @@ swaphead(int osvers, int osrel, int ossub)
 }
 
 static int
-swapline(struct sstat *ss, struct pstat *ps, int nproc,
+swapline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	unsigned int	badness;
 
@@ -1377,32 +1392,69 @@ swapline(struct sstat *ss, struct pstat *ps, int nproc,
 ** disk statistics
 */
 static void
-diskhead(int osvers, int osrel, int ossub)
+lvmhead(int osvers, int osrel, int ossub)
 {
-	printf("disk    busy   read/s KB/read   "
-	       "write/s KB/writ   avque avserv _disk_");
+	printf("disk           busy read/s KB/read  "
+	       "writ/s KB/writ avque avserv _lvm_");
+}
+
+static void
+mddhead(int osvers, int osrel, int ossub)
+{
+	printf("disk           busy read/s KB/read  "
+	       "writ/s KB/writ avque avserv _mdd_");
+}
+
+static void
+dskhead(int osvers, int osrel, int ossub)
+{
+	printf("disk           busy read/s KB/read  "
+	       "writ/s KB/writ avque avserv _dsk_");
 }
 
 static int
-diskline(struct sstat *ss, struct pstat *ps, int nproc,
-        time_t deltasec, time_t deltatic, time_t hz,
-        int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+gendskline(struct sstat *ss, char *tstamp, char selector)
 {
 	static char	firstcall = 1;
-	register int	i, nlines = 0;
+	register int	i, nlines = 0, nunit = 0;
 	count_t		mstot, iotot;
-	struct perxdsk 	*dp = ss->xdsk.xdsk;
+	struct perdsk 	*dp;
 	unsigned int	badness;
 
+	switch (selector)
+	{
+	   case 'l':
+		dp 	= ss->dsk.lvm;
+		nunit	= ss->dsk.nlvm;
+		break;
+
+	   case 'm':
+		dp 	= ss->dsk.mdd;
+		nunit	= ss->dsk.nmdd;
+		break;
+
+	   case 'd':
+		dp 	= ss->dsk.dsk;
+		nunit	= ss->dsk.ndsk;
+		break;
+
+	   default:
+		return 0;
+	}
+
         mstot  = (ss->cpu.all.stime + ss->cpu.all.utime +
                   ss->cpu.all.ntime + ss->cpu.all.itime +
                   ss->cpu.all.wtime + ss->cpu.all.Itime +
-                  ss->cpu.all.Stime + ss->cpu.all.steal  )
+                  ss->cpu.all.Stime + ss->cpu.all.steal +
+                  ss->cpu.all.guest                      )
 				* (count_t)1000 / hertz / ss->cpu.nrcpu;
 
-	for (i=0; i < ss->xdsk.nrxdsk; i++, dp++)
+	for (i=0; i < nunit; i++, dp++)
 	{
+		char	buf[32];
+		char	*pn;
+		int	len;
+
 		if ( (iotot = dp->nread + dp->nwrite) == 0 &&
  		     !firstcall && !allresources             )
 			continue;	/* no activity on this disk */
@@ -1414,16 +1466,21 @@ diskline(struct sstat *ss, struct pstat *ps, int nproc,
 			printf("%s  ", tstamp);
 
 		if (dskbadness)
-			badness = (ss->xdsk.xdsk[i].io_ms * 100.0 / mstot)
-                        		* 100 / dskbadness;
+			badness = (dp->io_ms * 100.0 / mstot) * 100/dskbadness;
                 else
 			badness = 0;
 
 		preprint(badness);
 
-		printf("%-7s %3.0lf%% %8.2lf %7.1lf %9.2lf %7.1lf "
-		       "%7.2lf %6.2lf ms",
-		    	dp->name,
+		if ( (len = strlen(dp->name)) > 14)
+			pn = dp->name + len - 14;
+		else
+			pn = dp->name;
+
+        sprintf(buf, "%12.12s", pn);
+		printf("%-14s %3.0lf%% %6.1lf %7.1lf %7.1lf %7.1lf "
+		       "%5.1lf %6.2lf ms",
+		    	pn,
 			mstot ? (double)dp->io_ms  *  100.0 / mstot   : 0.0,
 			mstot ? (double)dp->nread  * 1000.0 / mstot   : 0.0,
 			dp->nread  ?
@@ -1444,9 +1501,37 @@ diskline(struct sstat *ss, struct pstat *ps, int nproc,
 	}
 
 	firstcall = 0;
+
 	return nlines;
 }
 
+static int
+lvmline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
+        time_t deltasec, time_t deltatic, time_t hz,
+        int osvers, int osrel, int ossub, char *tstamp,
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
+{
+	return gendskline(ss, tstamp, 'l');
+}
+
+static int
+mddline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
+        time_t deltasec, time_t deltatic, time_t hz,
+        int osvers, int osrel, int ossub, char *tstamp,
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
+{
+	return gendskline(ss, tstamp, 'm');
+}
+
+static int
+dskline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
+        time_t deltasec, time_t deltatic, time_t hz,
+        int osvers, int osrel, int ossub, char *tstamp,
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
+{
+	return gendskline(ss, tstamp, 'd');
+}
+
 /*
 ** network-interface statistics
 */
@@ -1458,10 +1543,10 @@ ifhead(int osvers, int osrel, int ossub)
 }
 
 static int
-ifline(struct sstat *ss, struct pstat *ps, int nproc,
+ifline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	static char	firstcall = 1;
 	register long	i, nlines = 0;
@@ -1564,10 +1649,10 @@ IFhead(int osvers, int osrel, int ossub)
 }
 
 static int
-IFline(struct sstat *ss, struct pstat *ps, int nproc,
+IFline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	static char	firstcall = 1;
 	register long	i, nlines = 0;
@@ -1619,10 +1704,10 @@ ipv4head(int osvers, int osrel, int ossub)
 }
 
 static int
-ipv4line(struct sstat *ss, struct pstat *ps, int nproc,
+ipv4line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%8.1lf %8.1lf %11.1lf %9.1lf %9.1lf %11.1lf\n", 
 		(double)ss->net.ipv4.InReceives  / deltasec,
@@ -1642,10 +1727,10 @@ IPv4head(int osvers, int osrel, int ossub)
 }
 
 static int
-IPv4line(struct sstat *ss, struct pstat *ps, int nproc,
+IPv4line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("    %5.1lf %6.1lf %6.1lf %6.1lf %7.1lf %7.1lf  "
 	       "    %5.1lf %5.1lf\n", 
@@ -1671,10 +1756,10 @@ icmpv4head(int osvers, int osrel, int ossub)
 }
 
 static int
-icmpv4line(struct sstat *ss, struct pstat *ps, int nproc,
+icmpv4line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%7.1lf %8.1lf  %8.2lf %8.2lf  %8.2lf %8.2lf\n", 
 		(double)ss->net.icmpv4.InMsgs      / deltasec, 
@@ -1694,10 +1779,10 @@ ICMPv4head(int osvers, int osrel, int ossub)
 }
 
 static int
-ICMPv4line(struct sstat *ss, struct pstat *ps, int nproc,
+ICMPv4line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%6.2lf %5.2lf %5.2lf %5.2lf %5.2lf "
 	       "%6.2lf %5.2lf %5.2lf %5.2lf %5.2lf\n", 
@@ -1725,10 +1810,10 @@ udpv4head(int osvers, int osrel, int ossub)
 }
 
 static int
-udpv4line(struct sstat *ss, struct pstat *ps, int nproc,
+udpv4line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%9.1lf %10.1lf   %7.2lf %9.2lf\n",
 		(double)ss->net.udpv4.InDatagrams  / deltasec,
@@ -1749,10 +1834,10 @@ ipv6head(int osvers, int osrel, int ossub)
 }
 
 static int
-ipv6line(struct sstat *ss, struct pstat *ps, int nproc,
+ipv6line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%8.1lf %8.1lf %6.1lf %7.1lf %9.1lf %9.1lf %9.1lf\n", 
 		(double)ss->net.ipv6.Ip6InReceives   / deltasec,
@@ -1773,10 +1858,10 @@ IPv6head(int osvers, int osrel, int ossub)
 }
 
 static int
-IPv6line(struct sstat *ss, struct pstat *ps, int nproc,
+IPv6line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("    %5.1lf %6.1lf %6.1lf %6.1lf %7.1lf %7.1lf  "
 	       "    %5.1lf %5.1lf\n", 
@@ -1802,10 +1887,10 @@ icmpv6head(int osvers, int osrel, int ossub)
 }
 
 static int
-icmpv6line(struct sstat *ss, struct pstat *ps, int nproc,
+icmpv6line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%7.1lf %8.1lf %7.2lf %8.2lf %8.2lf %8.2lf %8.2lf\n", 
 		(double)ss->net.icmpv6.Icmp6InMsgs                  / deltasec, 
@@ -1827,10 +1912,10 @@ ICMPv6head(int osvers, int osrel, int ossub)
 }
 
 static int
-ICMPv6line(struct sstat *ss, struct pstat *ps, int nproc,
+ICMPv6line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%7.2lf %7.2lf %7.2lf %5.2lf %5.2lf "
 	       "%5.2lf %5.2lf %5.2lf %5.2lf\n", 
@@ -1857,10 +1942,10 @@ udpv6head(int osvers, int osrel, int ossub)
 }
 
 static int
-udpv6line(struct sstat *ss, struct pstat *ps, int nproc,
+udpv6line(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%9.1lf %10.1lf   %7.2lf %9.2lf\n",
 		(double)ss->net.udpv6.Udp6InDatagrams  / deltasec,
@@ -1881,10 +1966,10 @@ tcphead(int osvers, int osrel, int ossub)
 }
 
 static int
-tcpline(struct sstat *ss, struct pstat *ps, int nproc,
+tcpline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%8.1lf %9.1lf  %9.1lf %9.1lf  %7lld\n",
 		(double)ss->net.tcp.InSegs       / deltasec,
@@ -1903,10 +1988,10 @@ TCPhead(int osvers, int osrel, int ossub)
 }
 
 static int
-TCPline(struct sstat *ss, struct pstat *ps, int nproc,
+TCPline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%7.1lf  %9.1lf  %9.1lf  %12.1lf  %10.1lf\n",
 		(double)ss->net.tcp.InErrs       / deltasec,
@@ -1926,10 +2011,10 @@ httphead(int osvers, int osrel, int ossub)
 }
 
 static int
-httpline(struct sstat *ss, struct pstat *ps, int nproc,
+httpline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	printf("%10.2lf  %8.2lf  %9.2lf    %11d %11d\n",
 		(double)ss->www.accesses      / deltasec,
@@ -1954,14 +2039,14 @@ topchead(int osvers, int osrel, int ossub)
 }
 
 static int
-topcline(struct sstat *ss, struct pstat *ps, int nproc,
+topcline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
-	count_t		availcpu;
+	count_t	availcpu;
 
-	if (!ps)
+	if (!ts)
 	{
 		printf("report not available.....\n");
 		return 0;
@@ -1970,12 +2055,13 @@ topcline(struct sstat *ss, struct pstat *ps, int nproc,
 	/*
 	** sort process list in cpu order
 	*/
-	qsort(ps, nproc, sizeof(struct pstat), compcpu);
+	qsort(ps, nactproc, sizeof(struct tstat *), compcpu);
 
 	availcpu  = ss->cpu.all.stime + ss->cpu.all.utime +
 	            ss->cpu.all.ntime + ss->cpu.all.itime +
 	            ss->cpu.all.wtime + ss->cpu.all.Itime +
-		    ss->cpu.all.Stime + ss->cpu.all.steal; 
+		    ss->cpu.all.Stime + ss->cpu.all.steal +
+		    ss->cpu.all.guest; 
 
 	availcpu /= ss->cpu.nrcpu;
 
@@ -1984,12 +2070,12 @@ topcline(struct sstat *ss, struct pstat *ps, int nproc,
 
 	printf("%5d %-8.8s %3.0lf%% | %5d %-8.8s %3.0lf%% | "
 	       "%5d %-8.8s %3.0lf%%\n",
-		(ps+0)->gen.pid, (ps+0)->gen.name,
-		(double)((ps+0)->cpu.stime + (ps+0)->cpu.utime)*100.0/availcpu,
-		(ps+1)->gen.pid, (ps+1)->gen.name,
-		(double)((ps+1)->cpu.stime + (ps+1)->cpu.utime)*100.0/availcpu,
-		(ps+2)->gen.pid, (ps+2)->gen.name,
-		(double)((ps+2)->cpu.stime + (ps+2)->cpu.utime)*100.0/availcpu);
+	      (ps[0])->gen.pid, (ps[0])->gen.name,
+	      (double)((ps[0])->cpu.stime + (ps[0])->cpu.utime)*100.0/availcpu,
+	      (ps[1])->gen.pid, (ps[1])->gen.name,
+	      (double)((ps[1])->cpu.stime + (ps[1])->cpu.utime)*100.0/availcpu,
+	      (ps[2])->gen.pid, (ps[2])->gen.name,
+	      (double)((ps[2])->cpu.stime + (ps[2])->cpu.utime)*100.0/availcpu);
 
 	return 1;
 }
@@ -2005,14 +2091,14 @@ topmhead(int osvers, int osrel, int ossub)
 }
 
 static int
-topmline(struct sstat *ss, struct pstat *ps, int nproc,
+topmline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	count_t		availmem;
 
-	if (!ps)
+	if (!ts)
 	{
 		printf("report not available.....\n");
 		return 0;
@@ -2021,18 +2107,18 @@ topmline(struct sstat *ss, struct pstat *ps, int nproc,
 	/*
 	** sort process list in memory order
 	*/
-	qsort(ps, nproc, sizeof(struct pstat), compmem);
+	qsort(ps, nactproc, sizeof(struct tstat *), compmem);
 
 	availmem  = ss->mem.physmem * pagesize/1024;
 
 	printf("%5d %-8.8s %3.0lf%% | %5d %-8.8s %3.0lf%% | "
 	       "%5d %-8.8s %3.0lf%%\n",
-		(ps+0)->gen.pid, (ps+0)->gen.name,
-		(double)(ps+0)->mem.rmem * 100.0 / availmem,
-		(ps+1)->gen.pid, (ps+1)->gen.name,
-		(double)(ps+1)->mem.rmem * 100.0 / availmem,
-		(ps+2)->gen.pid, (ps+2)->gen.name,
-		(double)(ps+2)->mem.rmem * 100.0 / availmem);
+		(ps[0])->gen.pid, (ps[0])->gen.name,
+		(double)(ps[0])->mem.rmem * 100.0 / availmem,
+		(ps[1])->gen.pid, (ps[1])->gen.name,
+		(double)(ps[1])->mem.rmem * 100.0 / availmem,
+		(ps[2])->gen.pid, (ps[2])->gen.name,
+		(double)(ps[2])->mem.rmem * 100.0 / availmem);
 
 	return 1;
 }
@@ -2048,21 +2134,21 @@ topdhead(int osvers, int osrel, int ossub)
 }
 
 static int
-topdline(struct sstat *ss, struct pstat *ps, int nproc,
+topdline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	int		i;
 	count_t		availdsk;
 
-	if (!ps)
+	if (!ts)
 	{
 		printf("report not available.....\n");
 		return 0;
 	}
 
-	if ( !(supportflags & (PATCHSTAT | IOSTAT)) )
+	if ( !(supportflags & IOSTAT) )
 	{
 		printf("no per-process disk counters available.....\n");
 		return 0;
@@ -2071,9 +2157,9 @@ topdline(struct sstat *ss, struct pstat *ps, int nproc,
 	/*
 	** determine total disk accesses for all processes
 	*/
-	for (i=0, availdsk=0; i < nproc; i++)
+	for (i=0, availdsk=0; i < nactproc; i++)
 	{
-		availdsk += (ps+i)->dsk.rio + (ps+i)->dsk.wio;
+		availdsk += (ps[i])->dsk.rio + (ps[i])->dsk.wio;
 	}
 
 	if (availdsk == 0)
@@ -2082,16 +2168,16 @@ topdline(struct sstat *ss, struct pstat *ps, int nproc,
 	/*
 	** sort process list in disk order
 	*/
-	qsort(ps, nproc, sizeof(struct pstat), compdsk);
+	qsort(ps, nactproc, sizeof(struct tstat *), compdsk);
 
 	printf("%5d %-8.8s %3.0lf%% | %5d %-8.8s %3.0lf%% | "
 	       "%5d %-8.8s %3.0lf%%\n",
-		(ps+0)->gen.pid, (ps+0)->gen.name,
-	     	(double)((ps+0)->dsk.rio+(ps+0)->dsk.wio) *100.0/availdsk,
-		(ps+1)->gen.pid, (ps+1)->gen.name,
-		(double)((ps+1)->dsk.rio+(ps+1)->dsk.wio) *100.0/availdsk,
-		(ps+2)->gen.pid, (ps+2)->gen.name,
-		(double)((ps+2)->dsk.rio+(ps+2)->dsk.wio) *100.0/availdsk);
+		(ps[0])->gen.pid, (ps[0])->gen.name,
+	     	(double)((ps[0])->dsk.rio+(ps[0])->dsk.wio) *100.0/availdsk,
+		(ps[1])->gen.pid, (ps[1])->gen.name,
+		(double)((ps[1])->dsk.rio+(ps[1])->dsk.wio) *100.0/availdsk,
+		(ps[2])->gen.pid, (ps[2])->gen.name,
+		(double)((ps[2])->dsk.rio+(ps[2])->dsk.wio) *100.0/availdsk);
 
 	return 1;
 }
@@ -2107,21 +2193,21 @@ topnhead(int osvers, int osrel, int ossub)
 }
 
 static int
-topnline(struct sstat *ss, struct pstat *ps, int nproc,
+topnline(struct sstat *ss, struct tstat *ts, struct tstat **ps, int nactproc,
         time_t deltasec, time_t deltatic, time_t hz,
         int osvers, int osrel, int ossub, char *tstamp,
-        int ppres,  int pexit, int pzombie)
+        int ppres,  int ntrun, int ntslpi, int ntslpu, int pexit, int pzombie)
 {
 	int		i;
 	count_t		availnet;
 
-	if (!ps)
+	if (!ts)
 	{
 		printf("report not available.....\n");
 		return 0;
 	}
 
-	if ( !(supportflags & PATCHSTAT))
+	if ( !(supportflags & NETATOP) )
 	{
 		printf("no per-process network counters available.....\n");
 		return 0;
@@ -2130,11 +2216,10 @@ topnline(struct sstat *ss, struct pstat *ps, int nproc,
 	/*
 	** determine total network accesses for all processes
 	*/
-	for (i=0, availnet=0; i < nproc; i++)
+	for (i=0, availnet=0; i < nactproc; i++)
 	{
-		availnet += (ps+i)->net.tcpsnd + (ps+i)->net.tcprcv +
-		            (ps+i)->net.udpsnd + (ps+i)->net.udprcv +
-		            (ps+i)->net.rawsnd + (ps+i)->net.rawrcv;
+		availnet += (*(ps+i))->net.tcpssz + (*(ps+i))->net.tcprsz +
+		            (*(ps+i))->net.udpssz + (*(ps+i))->net.udprsz;
 	}
 
 	if (availnet == 0)
@@ -2143,26 +2228,22 @@ topnline(struct sstat *ss, struct pstat *ps, int nproc,
 	/*
 	** sort process list in network order
 	*/
-	qsort(ps, nproc, sizeof(struct pstat), compnet);
+	qsort(ps, nactproc, sizeof(struct tstat *), compnet);
 
 	printf("%5d %-8.8s %3.0lf%% | %5d %-8.8s %3.0lf%% | "
 	       "%5d %-8.8s %3.0lf%%\n",
-		(ps+0)->gen.pid, (ps+0)->gen.name,
-		(double)((ps+0)->net.tcpsnd + (ps+0)->net.tcprcv +
-		         (ps+0)->net.udpsnd + (ps+0)->net.udprcv +
-		         (ps+0)->net.rawsnd + (ps+0)->net.rawrcv)
+		(ps[0])->gen.pid, (ps[0])->gen.name,
+		(double)((ps[0])->net.tcpssz + (ps[0])->net.tcprsz +
+		         (ps[0])->net.udpssz + (ps[0])->net.udprsz  )
 							* 100.0 / availnet,
-		(ps+1)->gen.pid, (ps+1)->gen.name,
-		(double)((ps+1)->net.tcpsnd + (ps+1)->net.tcprcv +
-		         (ps+1)->net.udpsnd + (ps+1)->net.udprcv +
-		         (ps+1)->net.rawsnd + (ps+1)->net.rawrcv)
+		(ps[1])->gen.pid, (ps[1])->gen.name,
+		(double)((ps[1])->net.tcpssz + (ps[1])->net.tcprsz +
+		         (ps[1])->net.udpssz + (ps[1])->net.udprsz  )
 							* 100.0 / availnet,
-		(ps+2)->gen.pid, (ps+2)->gen.name,
-		(double)((ps+2)->net.tcpsnd + (ps+2)->net.tcprcv +
-		         (ps+2)->net.udpsnd + (ps+2)->net.udprcv +
-		         (ps+2)->net.rawsnd + (ps+2)->net.rawrcv)
+		(ps[2])->gen.pid, (ps[2])->gen.name,
+		(double)((ps[2])->net.tcpssz + (ps[2])->net.tcprsz +
+		         (ps[2])->net.udpssz + (ps[2])->net.udprsz  )
 							* 100.0 / availnet);
-
 	return 1;
 }
 
@@ -2209,7 +2290,9 @@ struct pridef pridef[] =
    {0,  "c",  'P',  taskhead,	taskline,  	"processes & threads",    },
    {0,  "m",  'm',  memhead,	memline,	"memory & swapspace",     },
    {0,  "m",  's',  swaphead,	swapline,	"swap rate",              },
-   {0,  "cd", 'd',  diskhead,	diskline,	"disk activity",          },
+   {0,  "cd", 'l',  lvmhead,	lvmline,	"logical volume activity", },
+   {0,  "cd", 'f',  mddhead,	mddline,	"multiple device activity",},
+   {0,  "cd", 'd',  dskhead,	dskline,	"disk activity",          },
    {0,  "n",  'i',  ifhead,	ifline,		"net-interf (general)",   },
    {0,  "n",  'I',  IFhead,	IFline,		"net-interf (errors)",    },
    {0,  "n",  'w',  ipv4head,	ipv4line,	"ip   v4    (general)",   },
