@@ -7,12 +7,12 @@
 ** This source-file contains functions to calculate the differences for
 ** the system-level and process-level counters since the previous sample.
 ** ==========================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        November 1996
 ** LINUX-port:  June 2000
 ** --------------------------------------------------------------------------
-** Copyright (C) 2000-2005 Gerlof Langeveld
+** Copyright (C) 2000-2010 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -30,6 +30,24 @@
 ** --------------------------------------------------------------------------
 **
 ** $Log: deviate.c,v $
+** Revision 1.45  2010/10/23 14:02:03  gerlof
+** Show counters for total number of running and sleep (S and D) threads.
+**
+** Revision 1.44  2010/05/18 19:19:43  gerlof
+** Introduce CPU frequency and scaling (JC van Winkel).
+**
+** Revision 1.43  2010/04/23 12:19:35  gerlof
+** Modified mail-address in header.
+**
+** Revision 1.42  2010/03/04 10:52:08  gerlof
+** Support I/O-statistics on logical volumes and MD devices.
+**
+** Revision 1.41  2009/12/31 11:34:21  gerlof
+** Sanity-check to bypass kernel-bug showing 497 days of CPU-consumption.
+**
+** Revision 1.40  2009/12/17 11:58:25  gerlof
+** Gather and display new counters: dirty cache and guest cpu usage.
+**
 ** Revision 1.39  2008/02/25 14:51:18  gerlof
 ** Experimental code for HTTP-statistics.
 **
@@ -150,7 +168,7 @@
 **
 */
 
-static const char rcsid[] = "$Id: deviate.c,v 1.39 2008/02/25 14:51:18 gerlof Exp $";
+static const char rcsid[] = "$Id: deviate.c,v 1.45 2010/10/23 14:02:03 gerlof Exp $";
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -172,155 +190,175 @@ static const char rcsid[] = "$Id: deviate.c,v 1.39 2008/02/25 14:51:18 gerlof Ex
 
 #define	MAX32BITVAL	0x100000000LL
 
+static void calcdiff(struct tstat *, struct tstat *, struct tstat *,
+							char, count_t);
+
 /*
 ** calculate the process-activity during the last sample
 */
 int
-deviatproc(struct pstat *aproc, int npresent,
-           struct pstat *eproc, int nexit, int deviatonly,
-	   struct pstat *dproc, int *nzombie)
+deviatproc(struct tstat *aproc, int npresent,
+           struct tstat *eproc, int nexit, int deviatonly,
+	   struct tstat *dproc, struct sstat *dstat, unsigned int *nactproc,
+           int *totproc, int *totrun, int *totslpi, int *totslpu, int *nzombie)
 {
 	register int		c, d;
-	register struct pstat	*curstat, *devstat;
-	struct pstat		prestat;
+	register struct tstat	*curstat, *devstat, *procstat = 0;
+	struct tstat		prestat;
 	struct pinfo		*pinfo;
+	count_t			totusedcpu;
+	char			procsaved = 1, hashtype = 'p';
 
 	/*
-	** make new list of all processes in the process-database;
-	** after handling all processes, the left-overs
-	** should have disappeared since the previous sample
+	** needed for sanity check later on...
+	*/
+	totusedcpu	= dstat->cpu.all.stime + dstat->cpu.all.utime +
+			  dstat->cpu.all.ntime + dstat->cpu.all.itime +
+			  dstat->cpu.all.wtime + dstat->cpu.all.Itime +
+			  dstat->cpu.all.Stime + dstat->cpu.all.steal +
+			  dstat->cpu.all.guest;
+
+	/*
+	** make new list of all tasks in the process-database;
+	** after handling all processes, the left-overs are processes
+	** that have disappeared since the previous sample
 	*/
 	pdb_makeresidue();
 
 	/*
 	** calculate deviations per present process
 	*/
-	for (c=0, d=0, *nzombie=0; c < npresent; c++)
+	*totproc=*totrun=*totslpi=*totslpu=*nzombie= 0;
+
+	for (c=0, d=0, *nactproc=0; c < npresent; c++)
 	{
-		char	newproc = 0;
+		char	newtask = 0;
 
 		curstat = aproc+c;
 
-		if (curstat->gen.state == 'Z')
-			(*nzombie)++;
+		if (curstat->gen.isproc)
+		{
+			(*totproc)++;
+
+			if (curstat->gen.state == 'Z')
+			{
+				(*nzombie)++;
+			}
+			else
+			{
+				*totrun		+= curstat->gen.nthrrun;
+				*totslpi	+= curstat->gen.nthrslpi;
+				*totslpu	+= curstat->gen.nthrslpu;
+			}
+		}
 
 		/*
-		** get previous figures from process-database
+		** get previous figures from task-database
 		*/
-		if ( pdb_getproc(curstat->gen.pid, curstat->gen.btime, &pinfo))
+		if ( pdb_gettask(curstat->gen.pid, curstat->gen.isproc,
+		                 curstat->gen.btime, &pinfo))
 		{
 			/*
-			** process already present during the previous sample;
-			** check differences with previous sample
+			** task already present during the previous sample;
+			** if no differences with previous sample, skip task
+			** unless all tasks have to be shown
+			**
+			** it might be that a process appears to have no
+			** differences while one its threads has differences;
+			** than the process will be inserted as well
 			*/
-			if (deviatonly && memcmp(curstat, &pinfo->pstat, 
-					           sizeof(struct pstat)) == EQ)
+			if (deviatonly && memcmp(curstat, &pinfo->tstat, 
+					           sizeof(struct tstat)) == EQ)
+			{
+				/* remember last unsaved process */
+				if (curstat->gen.isproc)
+				{
+					procstat  = curstat;
+					procsaved = 0;
+				}
+
 				continue;
+			}
 
 			/*
-			** differences detected, so the process was active,
+			** differences detected, so the task was active,
 		        ** or its status or memory-occupation has changed;
 			** save stats from previous sample (to use for
 			** further calculations) and store new statistics
-			** in process-database
+			** in task-database
 			*/
-			prestat 	= pinfo->pstat;	/* save old	*/
-			pinfo->pstat 	= *curstat;	/* overwrite	*/
+			prestat 	= pinfo->tstat;	/* save old	*/
+			pinfo->tstat 	= *curstat;	/* overwrite	*/
 		}
 		else
 		{
 			/*
-			** new process which must have been started during
+			** new task which must have been started during
 			** last interval
 			*/
 			memset(&prestat, 0, sizeof(prestat));
 
 			/*
-			** create new process
+			** create new task struct
 			*/
-			pdb_newproc(&pinfo);
+			pinfo = calloc(1, sizeof(struct pinfo));
 
-			pinfo->pstat = *curstat;
+			ptrverify(pinfo, "Malloc failed for new pinfo\n");
+
+			pinfo->tstat = *curstat;
 
 			/*
-			** add new process to process-database
+			** add new task to task-database
 			*/
-			pdb_addproc( curstat->gen.pid, pinfo);
+			pdb_addtask(curstat->gen.pid, pinfo);
 
-			newproc = 1;
+			newtask = 1;
 		}
 
 		/*
-		** now do the calculations
+		** active task found; do the difference calculations
 		*/
+		if (curstat->gen.isproc)
+		{
+			procsaved = 1;
+			(*nactproc)++;
+		}
+		else
+		{
+			/*
+			** active thread: check if related process registered
+			*/
+			if (!procsaved)
+			{
+				devstat = dproc+d;
+				calcdiff(devstat, procstat, procstat, 0,
+								totusedcpu);
+				procsaved = 1;
+				(*nactproc)++;
+				d++;
+			}
+		}
+
 		devstat = dproc+d;
 
-		devstat->gen        = curstat->gen;
-
-		if (newproc)
-			devstat->gen.excode |= ~(INT_MAX);
-
-		devstat->cpu.nice     = curstat->cpu.nice;
-		devstat->cpu.prio     = curstat->cpu.prio;
-		devstat->cpu.rtprio   = curstat->cpu.rtprio;
-		devstat->cpu.policy   = curstat->cpu.policy;
-		devstat->cpu.curcpu   = curstat->cpu.curcpu;
-		devstat->cpu.sleepavg = curstat->cpu.sleepavg;
-
-		devstat->cpu.stime  = 
-			subcount(curstat->cpu.stime, prestat.cpu.stime);
-		devstat->cpu.utime  =
-			subcount(curstat->cpu.utime, prestat.cpu.utime);
-
-		devstat->dsk.rio    =
-			subcount(curstat->dsk.rio, prestat.dsk.rio);
-		devstat->dsk.rsz    =
-			subcount(curstat->dsk.rsz, prestat.dsk.rsz);
-		devstat->dsk.wio    =
-			subcount(curstat->dsk.wio, prestat.dsk.wio);
-		devstat->dsk.wsz    =
-			subcount(curstat->dsk.wsz, prestat.dsk.wsz);
-		devstat->dsk.cwsz   =
-			subcount(curstat->dsk.cwsz, prestat.dsk.cwsz);
-
-		devstat->mem.shtext = curstat->mem.shtext;
-		devstat->mem.vmem   = curstat->mem.vmem;
-		devstat->mem.rmem   = curstat->mem.rmem;
-		devstat->mem.vgrow  = curstat->mem.vmem   - prestat.mem.vmem;
-		devstat->mem.rgrow  = curstat->mem.rmem   - prestat.mem.rmem;
-
-		devstat->mem.minflt = 
-			subcount(curstat->mem.minflt, prestat.mem.minflt);
-		devstat->mem.majflt =
-			subcount(curstat->mem.majflt, prestat.mem.majflt);
-
-		devstat->net.tcpsnd =
-			subcount(curstat->net.tcpsnd, prestat.net.tcpsnd);
-		devstat->net.tcpssz =
-			subcount(curstat->net.tcpssz, prestat.net.tcpssz);
-		devstat->net.tcprcv =
-			subcount(curstat->net.tcprcv, prestat.net.tcprcv);
-		devstat->net.tcprsz =
-			subcount(curstat->net.tcprsz, prestat.net.tcprsz);
-		devstat->net.udpsnd =
-			subcount(curstat->net.udpsnd, prestat.net.udpsnd);
-		devstat->net.udpssz =
-			subcount(curstat->net.udpssz, prestat.net.udpssz);
-		devstat->net.udprcv =
-			subcount(curstat->net.udprcv, prestat.net.udprcv);
-		devstat->net.udprsz =
-			subcount(curstat->net.udprsz, prestat.net.udprsz);
-		devstat->net.rawsnd =
-			subcount(curstat->net.rawsnd, prestat.net.rawsnd);
-		devstat->net.rawrcv =
-			subcount(curstat->net.rawrcv, prestat.net.rawrcv);
 
+		calcdiff(devstat, curstat, &prestat, newtask, totusedcpu);
 		d++;
 	}
 
 	/*
 	** calculate deviations per exited process
 	*/
+	if (nexit > 0 && supportflags&NETATOPD)
+	{
+		if (eproc->gen.pid)
+			hashtype = 'p';
+		else
+			hashtype = 'b';
+
+		netatop_exithash(hashtype);
+	}
+
 	for (c=0; c < nexit; c++)
 	{
 		/*
@@ -333,9 +371,9 @@ deviatproc(struct pstat *aproc, int npresent,
 
 		if (curstat->gen.pid)	/* acctrecord contains pid? */
 		{
-			if ( pdb_getproc(curstat->gen.pid,
+			if ( pdb_gettask(curstat->gen.pid, 1,
 			                 curstat->gen.btime, &pinfo))
-					prestat = pinfo->pstat;
+					prestat = pinfo->tstat;
 				else
 					memset(&prestat, 0, sizeof(prestat));
 		}
@@ -356,7 +394,7 @@ deviatproc(struct pstat *aproc, int npresent,
 				** against this exited one
 				*/
 				if ( pdb_srchresidue(curstat, &pinfo) )
-					prestat = pinfo->pstat;
+					prestat = pinfo->tstat;
 				else
 					memset(&prestat, 0, sizeof(prestat));
 		 	}
@@ -366,6 +404,7 @@ deviatproc(struct pstat *aproc, int npresent,
 		** now do the calculations
 		*/
 		devstat = dproc+d;
+		memset(devstat, 0, sizeof *devstat);
 
 		devstat->gen        = curstat->gen;
 
@@ -377,81 +416,6 @@ deviatproc(struct pstat *aproc, int npresent,
 
 		strcpy(devstat->gen.cmdline, prestat.gen.cmdline);
 
-		devstat->cpu.nice     = 0;
-		devstat->cpu.prio     = 0;
-		devstat->cpu.rtprio   = 0;
-		devstat->cpu.policy   = 0;
-		devstat->cpu.curcpu   = 0;
-		devstat->cpu.sleepavg = 0;
-
-		if (supportflags & PATCHACCT)
-		{
-			devstat->net.tcpsnd = curstat->net.tcpsnd -
-			                      prestat.net.tcpsnd;
-			devstat->net.tcpssz = curstat->net.tcpssz -
-			                      prestat.net.tcpssz;
-			devstat->net.tcprcv = curstat->net.tcprcv -
-			                      prestat.net.tcprcv;
-			devstat->net.tcprsz = curstat->net.tcprsz -
-			                      prestat.net.tcprsz;
-			devstat->net.udpsnd = curstat->net.udpsnd -
-			                      prestat.net.udpsnd;
-			devstat->net.udpssz = curstat->net.udpssz -
-			                      prestat.net.udpssz;
-			devstat->net.udprcv = curstat->net.udprcv -
-			                      prestat.net.udprcv;
-			devstat->net.udprsz = curstat->net.udprsz -
-			                      prestat.net.udprsz;
-			devstat->net.rawsnd = curstat->net.rawsnd -
-			                      prestat.net.rawsnd;
-			devstat->net.rawrcv = curstat->net.rawrcv -
-			                      prestat.net.rawrcv;
-			devstat->dsk.rio    = curstat->dsk.rio    -
-			                      prestat.dsk.rio;
-			devstat->dsk.rsz    = curstat->dsk.rsz    -
-			                      prestat.dsk.rsz;
-			devstat->dsk.wio    = curstat->dsk.wio    -
-			                      prestat.dsk.wio;
-			devstat->dsk.wsz    = curstat->dsk.wsz    -
-			                      prestat.dsk.wsz;
-			devstat->dsk.cwsz   = curstat->dsk.cwsz    -
-			                      prestat.dsk.cwsz;
-			devstat->mem.vmem   = 0;
-			devstat->mem.rmem   = 0;
-			devstat->mem.vgrow  = -curstat->mem.vmem;
-			devstat->mem.rgrow  = -curstat->mem.rmem;
-		}
-		else
-		{
-			devstat->net.tcpsnd   = 0;
-			devstat->net.tcpssz   = 0;
-			devstat->net.tcprcv   = 0;
-			devstat->net.tcprsz   = 0;
-			devstat->net.udpsnd   = 0;
-			devstat->net.udpssz   = 0;
-			devstat->net.udprcv   = 0;
-			devstat->net.udprsz   = 0;
-			devstat->net.rawsnd   = 0;
-			devstat->net.rawrcv   = 0;
-			devstat->mem.vmem     = 0;
-			devstat->mem.rmem     = 0;
-			devstat->mem.vgrow    = 0;
-			devstat->mem.rgrow    = 0;
-			devstat->dsk.wio      = 0;
-			devstat->dsk.wsz      = 0;
-			devstat->dsk.cwsz     = 0;
-			devstat->dsk.rsz      = 0;
-			devstat->dsk.rio      = curstat->dsk.rio  -
-						prestat.dsk.rio   -
-						prestat.dsk.wio;
-		}
-
-		devstat->cpu.stime  = curstat->cpu.stime  - prestat.cpu.stime;
-		devstat->cpu.utime  = curstat->cpu.utime  - prestat.cpu.utime;
-		devstat->mem.minflt = curstat->mem.minflt - prestat.mem.minflt;
-		devstat->mem.majflt = curstat->mem.majflt - prestat.mem.majflt;
-		devstat->mem.shtext = 0;
-
 		/*
 		** due to the strange exponent-type storage of values
 		** in the process accounting record, the resource-value
@@ -459,45 +423,44 @@ deviatproc(struct pstat *aproc, int npresent,
 		** stored value of the last registered sample; in that
 		** case the deviation should be set to zero
 		*/
-		if (devstat->cpu.stime < 0)
-			devstat->cpu.stime = 0;
-		if (devstat->cpu.utime < 0)
-			devstat->cpu.utime = 0;
-		if (devstat->dsk.rio    < 0)
-			devstat->dsk.rio   = 0;
-		if (devstat->dsk.rsz    < 0)
-			devstat->dsk.rsz   = 0;
-		if (devstat->dsk.wio    < 0)
-			devstat->dsk.wio   = 0;
-		if (devstat->dsk.wsz    < 0)
-			devstat->dsk.wsz   = 0;
-		if (devstat->dsk.cwsz    < 0)
-			devstat->dsk.cwsz   = 0;
-		if (devstat->mem.minflt < 0)
-			devstat->mem.minflt = 0;
-		if (devstat->mem.majflt < 0)
-			devstat->mem.majflt = 0;
+		if (curstat->cpu.stime > prestat.cpu.stime)
+			devstat->cpu.stime  = curstat->cpu.stime -
+			                      prestat.cpu.stime;
 
-		if (supportflags & PATCHACCT)
+		if (curstat->cpu.utime > prestat.cpu.utime)
+			devstat->cpu.utime  = curstat->cpu.utime -
+			                      prestat.cpu.utime;
+
+		if (curstat->mem.minflt > prestat.mem.minflt)
+			devstat->mem.minflt = curstat->mem.minflt - 
+			                      prestat.mem.minflt;
+
+		if (curstat->mem.majflt > prestat.mem.majflt)
+			devstat->mem.majflt = curstat->mem.majflt -
+			                      prestat.mem.majflt;
+
+		if (curstat->dsk.rio > (prestat.dsk.rio + prestat.dsk.wio))
+			devstat->dsk.rio    = curstat->dsk.rio  -
+			                      prestat.dsk.rio   -
+			                      prestat.dsk.wio;
+
+		/*
+		** try to match the network counters of netatop
+		*/
+		if (supportflags & NETATOPD)
 		{
-			if (devstat->net.tcpsnd < 0)
-				 devstat->net.tcpsnd = 0;
-			if (devstat->net.tcprcv < 0)
-				 devstat->net.tcprcv = 0;
-			if (devstat->net.udpsnd < 0)
-				 devstat->net.udpsnd = 0;
-			if (devstat->net.udprcv < 0)
-				 devstat->net.udprcv = 0;
-			if (devstat->net.rawsnd < 0)
-				 devstat->net.rawsnd = 0;
-			if (devstat->net.rawrcv < 0)
-				 devstat->net.rawrcv = 0;
+			unsigned long	val = (hashtype == 'p' ?
+						curstat->gen.pid :
+						curstat->gen.btime);
+
+			netatop_exitfind(val, devstat, &prestat);
 		}
 
 		d++;
+		(*nactproc)++;
 
 		if (prestat.gen.pid > 0)
-			pdb_delproc(prestat.gen.pid);
+			pdb_deltask(prestat.gen.pid, prestat.gen.isproc);
 	}
 
 	/*
@@ -508,13 +471,132 @@ deviatproc(struct pstat *aproc, int npresent,
 	return d;
 }
 
+/*
+** calculate the differences between the current sample and
+** the previous sample for a task
+*/
+static void
+calcdiff(struct tstat *devstat, struct tstat *curstat, struct tstat *prestat,
+	                                      char newtask, count_t totusedcpu)
+{
+	devstat->gen          = curstat->gen;
+
+	if (newtask)
+		devstat->gen.excode |= ~(INT_MAX);
+
+	devstat->cpu.nice     = curstat->cpu.nice;
+	devstat->cpu.prio     = curstat->cpu.prio;
+	devstat->cpu.rtprio   = curstat->cpu.rtprio;
+	devstat->cpu.policy   = curstat->cpu.policy;
+	devstat->cpu.curcpu   = curstat->cpu.curcpu;
+	devstat->cpu.sleepavg = curstat->cpu.sleepavg;
+
+	devstat->cpu.stime  = 
+		subcount(curstat->cpu.stime, prestat->cpu.stime);
+	devstat->cpu.utime  =
+		subcount(curstat->cpu.utime, prestat->cpu.utime);
+
+	/*
+	** particular kernel versions sometimes supply a smaller
+	** amount for consumed CPU-ticks than a previous sample;
+	** with unsigned calculations this results in 497 days of
+	** CPU-consumption so a sanity-check is needed here...
+	*/
+	if (devstat->cpu.stime > totusedcpu)
+		devstat->cpu.stime = 1;
+
+	if (devstat->cpu.utime > totusedcpu)
+		devstat->cpu.utime = 1;
+
+	/*
+	** do further calculations
+	*/
+	devstat->dsk.rio    =
+		subcount(curstat->dsk.rio, prestat->dsk.rio);
+	devstat->dsk.rsz    =
+		subcount(curstat->dsk.rsz, prestat->dsk.rsz);
+	devstat->dsk.wio    =
+		subcount(curstat->dsk.wio, prestat->dsk.wio);
+	devstat->dsk.wsz    =
+		subcount(curstat->dsk.wsz, prestat->dsk.wsz);
+	devstat->dsk.cwsz   =
+		subcount(curstat->dsk.cwsz, prestat->dsk.cwsz);
+
+	devstat->mem.vexec  = curstat->mem.vexec;
+	devstat->mem.vmem   = curstat->mem.vmem;
+	devstat->mem.rmem   = curstat->mem.rmem;
+	devstat->mem.vgrow  = curstat->mem.vmem   - prestat->mem.vmem;
+	devstat->mem.rgrow  = curstat->mem.rmem   - prestat->mem.rmem;
+	devstat->mem.vdata  = curstat->mem.vdata;
+	devstat->mem.vstack = curstat->mem.vstack;
+	devstat->mem.vlibs  = curstat->mem.vlibs;
+	devstat->mem.vswap  = curstat->mem.vswap;
+
+	devstat->mem.minflt = 
+		subcount(curstat->mem.minflt, prestat->mem.minflt);
+	devstat->mem.majflt =
+		subcount(curstat->mem.majflt, prestat->mem.majflt);
+
+	/*
+ 	** network counters: due to an unload/load of the netatop module,
+	** previous counters might be larger than the current
+	*/
+	if (curstat->net.tcpsnd >= prestat->net.tcpsnd)
+		devstat->net.tcpsnd =
+			subcount(curstat->net.tcpsnd, prestat->net.tcpsnd);
+	else
+		devstat->net.tcpsnd = curstat->net.tcpsnd;
+
+	if (curstat->net.tcpssz >= prestat->net.tcpssz)
+		devstat->net.tcpssz =
+			subcount(curstat->net.tcpssz, prestat->net.tcpssz);
+	else
+		devstat->net.tcpssz = curstat->net.tcpssz;
+
+	if (curstat->net.tcprcv >= prestat->net.tcprcv)
+		devstat->net.tcprcv =
+			subcount(curstat->net.tcprcv, prestat->net.tcprcv);
+	else
+		devstat->net.tcprcv = curstat->net.tcprcv;
+
+	if (curstat->net.tcprsz >= prestat->net.tcprsz)
+		devstat->net.tcprsz =
+			subcount(curstat->net.tcprsz, prestat->net.tcprsz);
+	else
+		devstat->net.tcprsz = curstat->net.tcprsz;
+
+	if (curstat->net.udpsnd >= prestat->net.udpsnd)
+		devstat->net.udpsnd =
+			subcount(curstat->net.udpsnd, prestat->net.udpsnd);
+	else
+		devstat->net.udpsnd = curstat->net.udpsnd;
+
+	if (curstat->net.udpssz >= prestat->net.udpssz)
+		devstat->net.udpssz =
+			subcount(curstat->net.udpssz, prestat->net.udpssz);
+	else
+		devstat->net.udpssz = curstat->net.udpssz;
+
+	if (curstat->net.udprcv >= prestat->net.udprcv)
+		devstat->net.udprcv =
+			subcount(curstat->net.udprcv, prestat->net.udprcv);
+	else
+		devstat->net.udprcv = curstat->net.udprcv;
+
+	if (curstat->net.udprsz >= prestat->net.udprsz)
+		devstat->net.udprsz =
+			subcount(curstat->net.udprsz, prestat->net.udprsz);
+	else
+		devstat->net.udprsz = curstat->net.udprsz;
+}
+
 /*
 ** calculate the system-activity during the last sample
 */
 void
 deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev)
 {
-	register int	i;
+	register int	i, j;
 	count_t		*cdev, *ccur, *cpre;
 
 	dev->cpu.nrcpu     = cur->cpu.nrcpu;
@@ -531,34 +613,46 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev)
 	dev->cpu.all.Stime = subcount(cur->cpu.all.Stime, pre->cpu.all.Stime);
 
 	dev->cpu.all.steal = subcount(cur->cpu.all.steal, pre->cpu.all.steal);
+	dev->cpu.all.guest = subcount(cur->cpu.all.guest, pre->cpu.all.guest);
 
-	if (dev->cpu.nrcpu == 1)
+	for (i=0; i < dev->cpu.nrcpu; i++)
 	{
-		dev->cpu.cpu[0] = dev->cpu.all;
-	}
-	else
-	{
-		for (i=0; i < dev->cpu.nrcpu; i++)
-		{
-			dev->cpu.cpu[i].cpunr = cur->cpu.cpu[i].cpunr;
-			dev->cpu.cpu[i].stime = subcount(cur->cpu.cpu[i].stime,
-						         pre->cpu.cpu[i].stime);
-			dev->cpu.cpu[i].utime = subcount(cur->cpu.cpu[i].utime,
-					 	         pre->cpu.cpu[i].utime);
-			dev->cpu.cpu[i].ntime = subcount(cur->cpu.cpu[i].ntime,
-						         pre->cpu.cpu[i].ntime);
-			dev->cpu.cpu[i].itime = subcount(cur->cpu.cpu[i].itime,
-						         pre->cpu.cpu[i].itime);
-			dev->cpu.cpu[i].wtime = subcount(cur->cpu.cpu[i].wtime,
-						         pre->cpu.cpu[i].wtime);
-			dev->cpu.cpu[i].Itime = subcount(cur->cpu.cpu[i].Itime,
-						         pre->cpu.cpu[i].Itime);
-			dev->cpu.cpu[i].Stime = subcount(cur->cpu.cpu[i].Stime,
-						         pre->cpu.cpu[i].Stime);
+		count_t 	ticks;
 
-			dev->cpu.cpu[i].steal = subcount(cur->cpu.cpu[i].steal,
-						         pre->cpu.cpu[i].steal);
-		}
+		dev->cpu.cpu[i].cpunr = cur->cpu.cpu[i].cpunr;
+		dev->cpu.cpu[i].stime = subcount(cur->cpu.cpu[i].stime,
+					         pre->cpu.cpu[i].stime);
+		dev->cpu.cpu[i].utime = subcount(cur->cpu.cpu[i].utime,
+				 	         pre->cpu.cpu[i].utime);
+		dev->cpu.cpu[i].ntime = subcount(cur->cpu.cpu[i].ntime,
+					         pre->cpu.cpu[i].ntime);
+		dev->cpu.cpu[i].itime = subcount(cur->cpu.cpu[i].itime,
+					         pre->cpu.cpu[i].itime);
+		dev->cpu.cpu[i].wtime = subcount(cur->cpu.cpu[i].wtime,
+					         pre->cpu.cpu[i].wtime);
+		dev->cpu.cpu[i].Itime = subcount(cur->cpu.cpu[i].Itime,
+					         pre->cpu.cpu[i].Itime);
+		dev->cpu.cpu[i].Stime = subcount(cur->cpu.cpu[i].Stime,
+					         pre->cpu.cpu[i].Stime);
+
+		dev->cpu.cpu[i].steal = subcount(cur->cpu.cpu[i].steal,
+					         pre->cpu.cpu[i].steal);
+		dev->cpu.cpu[i].guest = subcount(cur->cpu.cpu[i].guest,
+					         pre->cpu.cpu[i].guest);
+
+		ticks 		      = cur->cpu.cpu[i].freqcnt.ticks;
+
+		dev->cpu.cpu[i].freqcnt.maxfreq = 
+					cur->cpu.cpu[i].freqcnt.maxfreq;
+		dev->cpu.cpu[i].freqcnt.cnt = ticks ?
+					subcount(cur->cpu.cpu[i].freqcnt.cnt,
+					         pre->cpu.cpu[i].freqcnt.cnt)
+					       : cur->cpu.cpu[i].freqcnt.cnt;
+
+		dev->cpu.cpu[i].freqcnt.ticks = ticks ?
+					subcount(cur->cpu.cpu[i].freqcnt.ticks,
+					         pre->cpu.cpu[i].freqcnt.ticks)
+					       : cur->cpu.cpu[i].freqcnt.ticks;
 	}
 
 	dev->cpu.lavg1		= cur->cpu.lavg1;
@@ -569,21 +663,28 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev)
 	dev->mem.freemem	= cur->mem.freemem;
 	dev->mem.buffermem	= cur->mem.buffermem;
 	dev->mem.slabmem	= cur->mem.slabmem;
+	dev->mem.slabreclaim	= cur->mem.slabreclaim;
 	dev->mem.committed	= cur->mem.committed;
 	dev->mem.commitlim	= cur->mem.commitlim;
 	dev->mem.cachemem	= cur->mem.cachemem;
+	dev->mem.cachedrt	= cur->mem.cachedrt;
 	dev->mem.totswap	= cur->mem.totswap;
 	dev->mem.freeswap	= cur->mem.freeswap;
 
+	dev->mem.shmem		= cur->mem.shmem;
+	dev->mem.shmrss		= cur->mem.shmrss;
+	dev->mem.shmswp		= cur->mem.shmswp;
+
 	dev->mem.swouts		= subcount(cur->mem.swouts,  pre->mem.swouts);
 	dev->mem.swins		= subcount(cur->mem.swins,   pre->mem.swins);
 	dev->mem.pgscans	= subcount(cur->mem.pgscans, pre->mem.pgscans);
+	dev->mem.pgsteal	= subcount(cur->mem.pgsteal, pre->mem.pgsteal);
 	dev->mem.allocstall	= subcount(cur->mem.allocstall,
 				                         pre->mem.allocstall);
 
 	/*
 	** structures with network-related counters are considered
-	** as tabels of frequency-counters that have to be subtracte;
+	** as tables of frequency-counters that have to be subtracted;
 	** values that do not represent a frequency are corrected afterwards
 	*/
 	for (cdev = (count_t *)&dev->net.ipv4,
@@ -698,11 +799,12 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev)
 			** take care that interface properties are
 			** corrected for future samples
 			*/
-                        seteuid(0);		/* get root privileges      */
+                        regainrootprivs();	/* get root privileges      */
 
 		        initifprop();		/* refresh interface info   */
 
-                        seteuid( getuid() );	/* release root privileges  */
+			if (! droprootprivs())  /* drop setuid-root privs   */
+				cleanstop(42);
 
 			for (j=0; cur->intf.intf[j].name[0]; j++)
 			{
@@ -772,26 +874,170 @@ deviatsyst(struct sstat *cur, struct sstat *pre, struct sstat *dev)
 	/*
 	** calculate deviations for disks
 	*/
-	for (i=0; cur->xdsk.xdsk[i].name[0]; i++)
+	for (i=j=0; cur->dsk.dsk[i].name[0]; i++)
 	{
-		strcpy(dev->xdsk.xdsk[i].name, cur->xdsk.xdsk[i].name);
+		int	realj = j;
 
-		dev->xdsk.xdsk[i].nread  = subcount(cur->xdsk.xdsk[i].nread,
-		                                    pre->xdsk.xdsk[i].nread);
-		dev->xdsk.xdsk[i].nwrite = subcount(cur->xdsk.xdsk[i].nwrite,
-		                                    pre->xdsk.xdsk[i].nwrite);
-		dev->xdsk.xdsk[i].nrsect = subcount(cur->xdsk.xdsk[i].nrsect,
-		                                    pre->xdsk.xdsk[i].nrsect);
-		dev->xdsk.xdsk[i].nwsect = subcount(cur->xdsk.xdsk[i].nwsect,
-		                                    pre->xdsk.xdsk[i].nwsect);
-		dev->xdsk.xdsk[i].io_ms  = subcount(cur->xdsk.xdsk[i].io_ms,
-		                                    pre->xdsk.xdsk[i].io_ms);
-		dev->xdsk.xdsk[i].avque  = subcount(cur->xdsk.xdsk[i].avque,
-		                                    pre->xdsk.xdsk[i].avque);
+		/*
+ 		** check if disk has been added or removed since
+		** previous interval
+		*/
+		if ( strcmp(cur->dsk.dsk[i].name, pre->dsk.dsk[j].name) != 0)
+		{
+			for (j++; pre->dsk.dsk[j].name[0]; j++)
+			{
+				if ( strcmp(cur->dsk.dsk[i].name,
+						pre->dsk.dsk[j].name) == 0)
+					break;
+			}
+
+			/*
+			** either the corresponding entry has been found
+			** in the case that a disk has been removed, or
+			** an empty entry has been found (all counters
+			** on zero) in the case that a disk has been added
+			** during the last sample
+			*/
+		}
+
+		strcpy(dev->dsk.dsk[i].name, cur->dsk.dsk[i].name);
+
+		dev->dsk.dsk[i].nread  = subcount(cur->dsk.dsk[i].nread,
+		                                  pre->dsk.dsk[j].nread);
+		dev->dsk.dsk[i].nwrite = subcount(cur->dsk.dsk[i].nwrite,
+		                                  pre->dsk.dsk[j].nwrite);
+		dev->dsk.dsk[i].nrsect = subcount(cur->dsk.dsk[i].nrsect,
+		                                  pre->dsk.dsk[j].nrsect);
+		dev->dsk.dsk[i].nwsect = subcount(cur->dsk.dsk[i].nwsect,
+		                                  pre->dsk.dsk[j].nwsect);
+		dev->dsk.dsk[i].io_ms  = subcount(cur->dsk.dsk[i].io_ms,
+		                                  pre->dsk.dsk[j].io_ms);
+		dev->dsk.dsk[i].avque  = subcount(cur->dsk.dsk[i].avque,
+		                                  pre->dsk.dsk[j].avque);
+
+		/*
+		** determine new j
+		*/
+		if (pre->dsk.dsk[j].name)	// existing matching entry
+			j++;
+		else
+			j = realj;		// empty entry: stick to old j
 	}
 
-	dev->xdsk.xdsk[i].name[0] = '\0';
-	dev->xdsk.nrxdsk = i;
+	dev->dsk.dsk[i].name[0] = '\0';
+	dev->dsk.ndsk = i;
+
+	/*
+	** calculate deviations for multiple devices
+	*/
+	for (i=j=0; cur->dsk.mdd[i].name[0]; i++)
+	{
+		int	realj = j;
+
+		/*
+ 		** check if md has been added or removed since
+		** previous interval
+		*/
+		if ( strcmp(cur->dsk.mdd[i].name, pre->dsk.mdd[j].name) != 0)
+		{
+			for (j++; pre->dsk.mdd[j].name[0]; j++)
+			{
+				if ( strcmp(cur->dsk.mdd[i].name,
+						pre->dsk.mdd[j].name) == 0)
+					break;
+			}
+
+			/*
+			** either the corresponding entry has been found
+			** in the case that a md has been removed, or
+			** an empty entry has been found (all counters
+			** on zero) in the case that a md has been added
+			** during the last sample
+			*/
+		}
+
+		strcpy(dev->dsk.mdd[i].name, cur->dsk.mdd[i].name);
+
+		dev->dsk.mdd[i].nread  = subcount(cur->dsk.mdd[i].nread,
+		                                  pre->dsk.mdd[j].nread);
+		dev->dsk.mdd[i].nwrite = subcount(cur->dsk.mdd[i].nwrite,
+		                                  pre->dsk.mdd[j].nwrite);
+		dev->dsk.mdd[i].nrsect = subcount(cur->dsk.mdd[i].nrsect,
+		                                  pre->dsk.mdd[j].nrsect);
+		dev->dsk.mdd[i].nwsect = subcount(cur->dsk.mdd[i].nwsect,
+		                                  pre->dsk.mdd[j].nwsect);
+		dev->dsk.mdd[i].io_ms  = subcount(cur->dsk.mdd[i].io_ms,
+		                                  pre->dsk.mdd[j].io_ms);
+		dev->dsk.mdd[i].avque  = subcount(cur->dsk.mdd[i].avque,
+		                                  pre->dsk.mdd[j].avque);
+
+		/*
+		** determine new j
+		*/
+		if (pre->dsk.mdd[j].name)	// existing matching entry
+			j++;
+		else
+			j = realj;		// empty entry: stick to old j
+	}
+
+	dev->dsk.mdd[i].name[0] = '\0';
+	dev->dsk.nmdd = i;
+
+	/*
+	** calculate deviations for LVM logical volumes
+	*/
+	for (i=j=0; cur->dsk.lvm[i].name[0]; i++)
+	{
+		int	realj = j;
+
+		/*
+ 		** check if logical volume has been added or removed since
+		** previous interval
+		*/
+		if ( strcmp(cur->dsk.lvm[i].name, pre->dsk.lvm[j].name) != 0)
+		{
+			for (j++; pre->dsk.lvm[j].name[0]; j++)
+			{
+				if ( strcmp(cur->dsk.lvm[i].name,
+						pre->dsk.lvm[j].name) == 0)
+					break;
+			}
+
+			/*
+			** either the corresponding entry has been found
+			** in the case that a logical volume has been removed,
+			** or an empty entry has been found (all counters
+			** on zero) in the case that a logical volume has
+			** been added during the last sample
+			*/
+		}
+
+		strcpy(dev->dsk.lvm[i].name, cur->dsk.lvm[i].name);
+
+		dev->dsk.lvm[i].nread  = subcount(cur->dsk.lvm[i].nread,
+		                                  pre->dsk.lvm[j].nread);
+		dev->dsk.lvm[i].nwrite = subcount(cur->dsk.lvm[i].nwrite,
+		                                  pre->dsk.lvm[j].nwrite);
+		dev->dsk.lvm[i].nrsect = subcount(cur->dsk.lvm[i].nrsect,
+		                                  pre->dsk.lvm[j].nrsect);
+		dev->dsk.lvm[i].nwsect = subcount(cur->dsk.lvm[i].nwsect,
+		                                  pre->dsk.lvm[j].nwsect);
+		dev->dsk.lvm[i].io_ms  = subcount(cur->dsk.lvm[i].io_ms,
+		                                  pre->dsk.lvm[j].io_ms);
+		dev->dsk.lvm[i].avque  = subcount(cur->dsk.lvm[i].avque,
+		                                  pre->dsk.lvm[j].avque);
+
+		/*
+		** determine new j
+		*/
+		if (pre->dsk.lvm[j].name)	// existing matching entry
+			j++;
+		else
+			j = realj;		// empty entry: stick to old j
+	}
+
+	dev->dsk.lvm[i].name[0] = '\0';
+	dev->dsk.nlvm = i;
 
 	/*
 	** application-specific counters
@@ -841,6 +1087,7 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		tot->cpu.all.Itime += new->cpu.all.Itime;
 		tot->cpu.all.Stime += new->cpu.all.Stime;
 		tot->cpu.all.steal += new->cpu.all.steal;
+		tot->cpu.all.guest += new->cpu.all.guest;
 
 		if (new->cpu.nrcpu == 1)
 		{
@@ -859,6 +1106,7 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 				tot->cpu.cpu[i].Itime += new->cpu.cpu[i].Itime;
 				tot->cpu.cpu[i].Stime += new->cpu.cpu[i].Stime;
 				tot->cpu.cpu[i].steal += new->cpu.cpu[i].steal;
+				tot->cpu.cpu[i].guest += new->cpu.cpu[i].guest;
 			}
 		}
 
@@ -872,12 +1120,18 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		tot->mem.freemem	 = new->mem.freemem;
 		tot->mem.buffermem	 = new->mem.buffermem;
 		tot->mem.slabmem	 = new->mem.slabmem;
+		tot->mem.slabreclaim	 = new->mem.slabreclaim;
 		tot->mem.committed	 = new->mem.committed;
 		tot->mem.commitlim	 = new->mem.commitlim;
 		tot->mem.cachemem	 = new->mem.cachemem;
+		tot->mem.cachedrt	 = new->mem.cachedrt;
 		tot->mem.totswap	 = new->mem.totswap;
 		tot->mem.freeswap	 = new->mem.freeswap;
 
+		tot->mem.shmem		 = new->mem.shmem;
+		tot->mem.shmrss		 = new->mem.shmrss;
+		tot->mem.shmswp		 = new->mem.shmswp;
+
 		tot->mem.swouts		+= new->mem.swouts;
 		tot->mem.swins		+= new->mem.swins;
 		tot->mem.pgscans	+= new->mem.pgscans;
@@ -887,7 +1141,7 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 	   case 'n':	/* accumulate network-related counters */
 		/*
 		** structures with network-related counters are considered
-		** as tabels of frequency-counters that will be accumulated;
+		** as tables of frequency-counters that will be accumulated;
 		** values that do not represent a frequency are corrected
 		** afterwards
 		*/
@@ -1021,20 +1275,50 @@ totalsyst(char category, struct sstat *new, struct sstat *tot)
 		break;
 
 	   case 'd':	/* accumulate disk-related counters */
-		for (i=0; new->xdsk.xdsk[i].name[0]; i++)
+		for (i=0; new->dsk.dsk[i].name[0]; i++)
 		{
-			strcpy(tot->xdsk.xdsk[i].name, new->xdsk.xdsk[i].name);
+			strcpy(tot->dsk.dsk[i].name, new->dsk.dsk[i].name);
 	
-			tot->xdsk.xdsk[i].nread  += new->xdsk.xdsk[i].nread;
-			tot->xdsk.xdsk[i].nwrite += new->xdsk.xdsk[i].nwrite;
-			tot->xdsk.xdsk[i].nrsect += new->xdsk.xdsk[i].nrsect;
-			tot->xdsk.xdsk[i].nwsect += new->xdsk.xdsk[i].nwsect;
-			tot->xdsk.xdsk[i].io_ms  += new->xdsk.xdsk[i].io_ms;
-			tot->xdsk.xdsk[i].avque  += new->xdsk.xdsk[i].avque;
+			tot->dsk.dsk[i].nread  += new->dsk.dsk[i].nread;
+			tot->dsk.dsk[i].nwrite += new->dsk.dsk[i].nwrite;
+			tot->dsk.dsk[i].nrsect += new->dsk.dsk[i].nrsect;
+			tot->dsk.dsk[i].nwsect += new->dsk.dsk[i].nwsect;
+			tot->dsk.dsk[i].io_ms  += new->dsk.dsk[i].io_ms;
+			tot->dsk.dsk[i].avque  += new->dsk.dsk[i].avque;
 		}
 	
-		tot->xdsk.xdsk[i].name[0] = '\0';
-		tot->xdsk.nrxdsk = i;
+		tot->dsk.dsk[i].name[0] = '\0';
+		tot->dsk.ndsk = i;
+
+		for (i=0; new->dsk.lvm[i].name[0]; i++)
+		{
+			strcpy(tot->dsk.lvm[i].name, new->dsk.lvm[i].name);
+	
+			tot->dsk.lvm[i].nread  += new->dsk.lvm[i].nread;
+			tot->dsk.lvm[i].nwrite += new->dsk.lvm[i].nwrite;
+			tot->dsk.lvm[i].nrsect += new->dsk.lvm[i].nrsect;
+			tot->dsk.lvm[i].nwsect += new->dsk.lvm[i].nwsect;
+			tot->dsk.lvm[i].io_ms  += new->dsk.lvm[i].io_ms;
+			tot->dsk.lvm[i].avque  += new->dsk.lvm[i].avque;
+		}
+	
+		tot->dsk.lvm[i].name[0] = '\0';
+		tot->dsk.nlvm = i;
+
+		for (i=0; new->dsk.mdd[i].name[0]; i++)
+		{
+			strcpy(tot->dsk.mdd[i].name, new->dsk.mdd[i].name);
+	
+			tot->dsk.mdd[i].nread  += new->dsk.mdd[i].nread;
+			tot->dsk.mdd[i].nwrite += new->dsk.mdd[i].nwrite;
+			tot->dsk.mdd[i].nrsect += new->dsk.mdd[i].nrsect;
+			tot->dsk.mdd[i].nwsect += new->dsk.mdd[i].nwsect;
+			tot->dsk.mdd[i].io_ms  += new->dsk.mdd[i].io_ms;
+			tot->dsk.mdd[i].avque  += new->dsk.mdd[i].avque;
+		}
+	
+		tot->dsk.mdd[i].name[0] = '\0';
+		tot->dsk.nmdd = i;
 		break;
 	}
 }
