@@ -7,8 +7,8 @@
 ** Include-file describing process-level counters maintained and functions
 ** to access the process-database.
 ** ================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        November 1996
 ** LINUX-port:  June 2000
 **
@@ -24,29 +24,38 @@
 */
 
 #define	PNAMLEN		15
-#define	CMDLEN		68
+#define	CMDLEN		255
 
 /* 
 ** structure containing only relevant process-info extracted 
 ** from kernel's process-administration
 */
-struct pstat {
-	/* GENERAL PROCESS INFO 					*/
+struct tstat {
+	/* GENERAL TASK INFO 					*/
 	struct gen {
+		int	tgid;		/* threadgroup identification 	*/
 		int	pid;		/* process identification 	*/
-		int	ruid;		/* real user  identification 	*/
-		int	rgid;		/* real group identification 	*/
 		int	ppid;           /* parent process identification*/
+		int	ruid;		/* real  user  identification 	*/
+		int	euid;		/* eff.  user  identification 	*/
+		int	suid;		/* saved user  identification 	*/
+		int	fsuid;		/* fs    user  identification 	*/
+		int	rgid;		/* real  group identification 	*/
+		int	egid;		/* eff.  group identification 	*/
+		int	sgid;		/* saved group identification 	*/
+		int	fsgid;		/* fs    group identification 	*/
 		int	nthr;		/* number of threads in tgroup 	*/
 		char	name[PNAMLEN+1];/* process name string       	*/
+		char 	isproc;		/* boolean: process level?      */
 		char 	state;		/* process state ('E' = exited)	*/
 		int	excode;		/* process exit status		*/
 		time_t 	btime;		/* process start time (epoch)	*/
+		time_t 	elaps;		/* process elaps time (hertz)	*/
 		char	cmdline[CMDLEN+1];/* command-line string       	*/
 		int	nthrslpi;	/* # threads in state 'S'       */
 		int	nthrslpu;	/* # threads in state 'D'       */
 		int	nthrrun;	/* # threads in state 'R'       */
-		int	ifuture[1];     /* reserved                     */
+		int	ifuture[4];     /* reserved                     */
 	} gen;
 
 	/* CPU STATISTICS						*/
@@ -78,12 +87,15 @@ struct pstat {
 	struct mem {
 		count_t	minflt;		/* number of page-reclaims 	*/
 		count_t	majflt;		/* number of page-faults 	*/
-		count_t	shtext;		/* text     memory (Kb)         */
-		count_t	vmem;		/* virtual  memory (Kb)		*/
-		count_t	rmem;		/* resident memory (Kb)		*/
-		count_t vgrow;		/* virtual  growth (Kb)    	*/
-		count_t rgrow;		/* resident growth (Kb)     	*/
-		count_t	cfuture[4];	/* reserved for future use	*/
+		count_t	vexec;		/* virtmem execfile (Kb)        */
+		count_t	vmem;		/* virtual  memory  (Kb)	*/
+		count_t	rmem;		/* resident memory  (Kb)	*/
+		count_t vgrow;		/* virtual  growth  (Kb)    	*/
+		count_t rgrow;		/* resident growth  (Kb)     	*/
+		count_t vdata;		/* virtmem data     (Kb)     	*/
+		count_t vstack;		/* virtmem stack    (Kb)     	*/
+		count_t vlibs;		/* virtmem libexec  (Kb)     	*/
+		count_t vswap;		/* swap space used  (Kb)     	*/
 	} mem;
 
 	/* NETWORK STATISTICS						*/
@@ -96,8 +108,8 @@ struct pstat {
 		count_t udpssz;		/* cumulative size packets sent	*/
 		count_t	udprcv;		/* number of UDP-packets recved	*/
 		count_t udprsz;		/* cumulative size packets sent	*/
-		count_t	rawsnd;		/* number of raw packets sent	*/
-		count_t	rawrcv;		/* number of raw packets recved	*/
+		count_t	avail1;		/* */
+		count_t	avail2;		/* */
 		count_t	cfuture[4];	/* reserved for future use	*/
 	} net;
 };
@@ -107,24 +119,27 @@ struct pinfo {
 	struct pinfo	*prnext;	/* next process in residue chain */
 	struct pinfo	*prprev;	/* prev process in residue chain */
 
-	struct pstat	pstat;		/* per-process statistics        */
+	struct tstat	tstat;		/* per-process statistics        */
 };
 
 /*
 ** prototypes of process-database functions
 */
-int		pdb_getproc(int, time_t, struct pinfo **);
-void		pdb_addproc(int, struct pinfo *);
-int		pdb_delproc(int);
-int		pdb_newproc(struct pinfo **);
+int		pdb_gettask(int, char, time_t, struct pinfo **);
+void		pdb_addtask(int, struct pinfo *);
+int		pdb_deltask(int, char);
 int		pdb_makeresidue(void);
 int		pdb_cleanresidue(void);
-int		pdb_srchresidue(struct pstat *, struct pinfo **);
+int		pdb_srchresidue(struct tstat *, struct pinfo **);
 
 /*
 ** prototypes for raw process-statistics functions
 */
-int		deviatproc(struct pstat *, int, struct pstat *, int, int,
-					struct pstat *, int *);
-int		photoproc(struct pstat *, int);
+struct netpertask;
+
+int		deviatproc(struct tstat *, int, struct tstat *, int, int,
+				struct tstat *, struct sstat *, unsigned int *,
+				int *, int *, int *, int *, int *);
+
+int		photoproc(struct tstat *, int);
 unsigned int	countprocs(void);
