@@ -7,12 +7,12 @@
 ** This source-file contains functions to read all relevant system-level
 ** figures.
 ** ==========================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        November 1996
 ** LINUX-port:  June 2000
 ** --------------------------------------------------------------------------
-** Copyright (C) 2000-2005 Gerlof Langeveld
+** Copyright (C) 2000-2012 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -30,6 +30,31 @@
 ** --------------------------------------------------------------------------
 **
 ** $Log: photosyst.c,v $
+** Revision 1.38  2010/11/19 07:40:40  gerlof
+** Support of virtual disk vd... (kvm).
+**
+** Revision 1.37  2010/11/14 06:42:18  gerlof
+** After opening /proc/cpuinfo, the file descriptor was not closed any more.
+**
+** Revision 1.36  2010/10/23 14:09:50  gerlof
+** Add support for mmcblk disks (MMC/SD cardreaders)
+** Credits: Anssi Hannula
+**
+** Revision 1.35  2010/05/18 19:20:30  gerlof
+** Introduce CPU frequency and scaling (JC van Winkel).
+**
+** Revision 1.34  2010/04/23 12:19:35  gerlof
+** Modified mail-address in header.
+**
+** Revision 1.33  2010/03/04 10:58:05  gerlof
+** Added recognition of device-type /dev/fio...
+**
+** Revision 1.32  2010/03/04 10:52:47  gerlof
+** Support I/O-statistics on logical volumes and MD devices.
+**
+** Revision 1.31  2009/12/17 11:59:16  gerlof
+** Gather and display new counters: dirty cache and guest cpu usage.
+**
 ** Revision 1.30  2008/02/25 13:47:00  gerlof
 ** Experimental code for HTTP statistics.
 **
@@ -124,7 +149,7 @@
 **
 */
 
-static const char rcsid[] = "$Id: photosyst.c,v 1.30 2008/02/25 13:47:00 gerlof Exp $";
+static const char rcsid[] = "$Id: photosyst.c,v 1.38 2010/11/19 07:40:40 gerlof Exp $";
 
 #include <sys/types.h>
 #include <stdio.h>
@@ -132,22 +157,35 @@ static const char rcsid[] = "$Id: photosyst.c,v 1.30 2008/02/25 13:47:00 gerlof
 #include <unistd.h>
 #include <stdlib.h>
 #include <regex.h>
+#include <sys/stat.h>
 #include <sys/times.h>
 #include <sys/wait.h>
 #include <time.h>
 #include <signal.h>
 #include <string.h>
+#include <dirent.h>
 
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <netdb.h>
 
+// #define	_GNU_SOURCE
+#include <sys/ipc.h>
+#include <sys/shm.h>
+
 #include "atop.h"
 #include "photosyst.h"
 
 #define	MAXCNT	64
 
-static int	isrealdisk(char *, char *, int);
+/* return value of isdisk() */
+#define	NONTYPE	0
+#define	DSKTYPE	1
+#define	MDDTYPE	2
+#define	LVMTYPE	3
+
+static int	isdisk(unsigned int, unsigned int,
+			char *, struct perdsk *, int);
 
 static struct ipv6_stats	ipv6_tmp;
 static struct icmpv6_stats	icmpv6_tmp;
@@ -235,14 +273,19 @@ photosyst(struct sstat *si)
 	FILE 		*fp;
 	char		linebuf[1024], nam[64], origdir[1024];
 	static char	part_stats = 1; /* per-partition statistics ? */
+	unsigned int	major, minor;
+	struct shm_info	shminfo;
 #if	HTTPSTATS
 	static int	wwwvalid = 1;
 #endif
 
 	memset(si, 0, sizeof(struct sstat));
 
-	getcwd(origdir, sizeof origdir);
-	chdir("/proc");
+	if ( getcwd(origdir, sizeof origdir) == NULL)
+		cleanstop(53);
+
+	if ( chdir("/proc") == -1)
+		cleanstop(53);
 
 	/*
 	** gather various general statistics from the file /proc/stat and
@@ -279,6 +322,9 @@ photosyst(struct sstat *si)
 
 					if (nr > 8)	/* steal support */
 						si->cpu.all.steal = cnts[7];
+
+					if (nr > 9)	/* guest support */
+						si->cpu.all.guest = cnts[8];
 				}
 				continue;
 			}
@@ -287,6 +333,14 @@ photosyst(struct sstat *si)
 			{
 				i = atoi(&nam[3]);
 
+				if (i >= MAXCPU)
+				{
+					fprintf(stderr,
+						"cpu %s exceeds maximum of %d\n",
+						nam, MAXCPU);
+					continue;
+				}
+
 				si->cpu.cpu[i].cpunr	= i;
 				si->cpu.cpu[i].utime	= cnts[0];
 				si->cpu.cpu[i].ntime	= cnts[1];
@@ -301,6 +355,9 @@ photosyst(struct sstat *si)
 
 					if (nr > 8)	/* steal support */
 						si->cpu.cpu[i].steal = cnts[7];
+
+					if (nr > 9)	/* guest support */
+						si->cpu.cpu[i].guest = cnts[8];
 				}
 
 				si->cpu.nrcpu++;
@@ -359,16 +416,138 @@ photosyst(struct sstat *si)
 		fclose(fp);
 	}
 
+	/*
+	** gather frequency scaling info.
+        ** sources (in order of preference): 
+        **     /sys/devices/system/cpu/cpu0/cpufreq/stats/time_in_state
+        ** or
+        **     /sys/devices/system/cpu/cpu0/cpufreq/scaling_cur_freq
+        **     /sys/devices/system/cpu/cpu0/cpufreq/scaling_max_freq
+        **
+	** store them in binary form
+	*/
+        static char fn[80];
+        int didone=0;
+
+        for (i = 0; i < si->cpu.nrcpu; ++i)
+        {
+                long long f=0;
+
+                sprintf(fn,
+                   "/sys/devices/system/cpu/cpu%d/cpufreq/stats/time_in_state",
+                   i);
+
+                if ((fp=fopen(fn, "r")) != 0)
+                {
+                        long long hits=0;
+                        long long maxfreq=0;
+                        long long cnt=0;
+                        long long sum=0;
+
+                        while (fscanf(fp, "%lld %lld", &f, &cnt) == 2)
+                        {
+                                f	/= 1000;
+                                sum 	+= (f*cnt);
+                                hits	+= cnt;
+
+                                if (f > maxfreq)
+                        		maxfreq=f;
+                        }
+
+	                si->cpu.cpu[i].freqcnt.maxfreq	= maxfreq;
+	                si->cpu.cpu[i].freqcnt.cnt	= sum;
+	                si->cpu.cpu[i].freqcnt.ticks	= hits;
+
+                        fclose(fp);
+                        didone=1;
+                }
+		else
+		{    // governor statistics not available
+                     sprintf(fn,  
+                      "/sys/devices/system/cpu/cpu%d/cpufreq/cpuinfo_max_freq",
+		      i);
+
+                        if ((fp=fopen(fn, "r")) != 0)
+                        {
+                                if (fscanf(fp, "%lld", &f) == 1)
+                                {
+  					// convert KHz to MHz
+	                                si->cpu.cpu[i].freqcnt.maxfreq =f/1000; 
+                                }
+
+                                didone=1;
+                                fclose(fp);
+                        }
+                        else 
+                        {
+	                        si->cpu.cpu[i].freqcnt.maxfreq=0;
+                        }
+
+                       sprintf(fn,  
+                       "/sys/devices/system/cpu/cpu%d/cpufreq/scaling_cur_freq",
+		       i);
+        
+                        if ((fp=fopen(fn, "r")) != 0)
+                        {
+                                if (fscanf(fp, "%lld", &f) == 1)
+                                {
+   					// convert KHz to MHz
+                                        si->cpu.cpu[i].freqcnt.cnt   = f/1000;
+                                        si->cpu.cpu[i].freqcnt.ticks = 0;
+                                }
+
+                                fclose(fp);
+                                didone=1;
+                        }
+                        else
+                        {
+                                si->cpu.cpu[i].freqcnt.maxfreq	= 0;
+                                si->cpu.cpu[i].freqcnt.cnt	= 0;
+                                si->cpu.cpu[i].freqcnt.ticks 	= 0;
+                                break;    // use cpuinfo
+                        }
+                }
+        } // for all CPUs
+
+        if (!didone)     // did not get processor freq statistics.
+                         // use /proc/cpuinfo
+        {
+	        if ( (fp = fopen("cpuinfo", "r")) != NULL)
+                {
+                        // get information from the lines
+                        // processor\t: 0
+                        // cpu MHz\t\t: 800.000
+                        
+                        int cpuno=-1;
+
+		        while ( fgets(linebuf, sizeof(linebuf), fp) != NULL)
+                        {
+                                if (memcmp(linebuf, "processor", 9)== EQ)
+					sscanf(linebuf, "%*s %*s %d", &cpuno);
+
+                                if (memcmp(linebuf, "cpu MHz", 7) == EQ)
+				{
+                                        if (cpuno >= 0 && cpuno < si->cpu.nrcpu)
+					{
+						sscanf(linebuf,
+							"%*s %*s %*s %lld",
+                                	     		&(si->cpu.cpu[cpuno].freqcnt.cnt));
+					}
+                                }
+                        }
+
+			fclose(fp);
+                }
+
+        }
+
 	/*
 	** gather virtual memory statistics from the file /proc/vmstat and
 	** store them in binary form (>= kernel 2.6)
 	*/
 	if ( (fp = fopen("vmstat", "r")) != NULL)
 	{
-		int	nrfields = 9;	/* number of fields to be filled */
-
-		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL &&
-								nrfields > 0)
+		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL)
 		{
 			nr = sscanf(linebuf, "%s %lld", nam, &cnts[0]);
 
@@ -378,28 +557,30 @@ photosyst(struct sstat *si)
 			if ( strcmp("pswpin", nam) == EQ)
 			{
 				si->mem.swins   = cnts[0];
-				nrfields--;
 				continue;
 			}
 
 			if ( strcmp("pswpout", nam) == EQ)
 			{
 				si->mem.swouts  = cnts[0];
-				nrfields--;
-				continue;
-			}
-
-			if ( strcmp("allocstall", nam) == EQ)
-			{
-				si->mem.allocstall = cnts[0];
-				nrfields--;
 				continue;
 			}
 
 			if ( strncmp("pgscan_", nam, 7) == EQ)
 			{
 				si->mem.pgscans += cnts[0];
-				nrfields--;
+				continue;
+			}
+
+			if ( strncmp("pgsteal_", nam, 8) == EQ)
+			{
+				si->mem.pgsteal += cnts[0];
+				continue;
+			}
+
+			if ( strcmp("allocstall", nam) == EQ)
+			{
+				si->mem.allocstall = cnts[0];
 				continue;
 			}
 		}
@@ -423,13 +604,15 @@ photosyst(struct sstat *si)
 	si->mem.buffermem	= (count_t)-1;
 	si->mem.cachemem  	= (count_t)-1;
 	si->mem.slabmem		= (count_t) 0;
+	si->mem.slabreclaim	= (count_t) 0;
+	si->mem.shmem 		= (count_t) 0;
 	si->mem.totswap  	= (count_t)-1;
 	si->mem.freeswap 	= (count_t)-1;
 	si->mem.committed 	= (count_t) 0;
 
 	if ( (fp = fopen("meminfo", "r")) != NULL)
 	{
-		int	nrfields = 9;	/* number of fields to be filled */
+		int	nrfields = 12;	/* number of fields to be filled */
 
 		while ( fgets(linebuf, sizeof(linebuf), fp) != NULL && 
 								nrfields > 0)
@@ -467,6 +650,12 @@ photosyst(struct sstat *si)
 						nrfields--;
 					}
 				}
+			else	if (strcmp("Dirty:", nam) == EQ)
+				{
+					si->mem.cachedrt  =
+							cnts[0]*1024/pagesize;
+					nrfields--;
+				}
 			else	if (strcmp("MemTotal:", nam) == EQ)
 				{
 					if (si->mem.physmem  == (count_t)-1)
@@ -494,6 +683,11 @@ photosyst(struct sstat *si)
 						nrfields--;
 					}
 				}
+			else	if (strcmp("Shmem:", nam) == EQ)
+				{
+					si->mem.shmem = cnts[0]*1024/pagesize;
+					nrfields--;
+				}
 			else	if (strcmp("SwapTotal:", nam) == EQ)
 				{
 					if (si->mem.totswap  == (count_t)-1)
@@ -517,6 +711,12 @@ photosyst(struct sstat *si)
 					si->mem.slabmem = cnts[0]*1024/pagesize;
 					nrfields--;
 				}
+			else	if (strcmp("SReclaimable:", nam) == EQ)
+				{
+					si->mem.slabreclaim = cnts[0]*1024/
+								pagesize;
+					nrfields--;
+				}
 			else	if (strcmp("Committed_AS:", nam) == EQ)
 				{
 					si->mem.committed = cnts[0]*1024/
@@ -712,12 +912,12 @@ photosyst(struct sstat *si)
 			      "%*d %*d %*d %255s %lld %*d %lld %*d "
 			      "%lld %*d %lld %*d %*d %lld %lld",
 			        diskname,
-				&(si->xdsk.xdsk[i].nread),
-				&(si->xdsk.xdsk[i].nrsect),
-				&(si->xdsk.xdsk[i].nwrite),
-				&(si->xdsk.xdsk[i].nwsect),
-				&(si->xdsk.xdsk[i].io_ms),
-				&(si->xdsk.xdsk[i].avque) );
+				&(si->dsk.dsk[i].nread),
+				&(si->dsk.dsk[i].nrsect),
+				&(si->dsk.dsk[i].nwrite),
+				&(si->dsk.dsk[i].nwsect),
+				&(si->dsk.dsk[i].io_ms),
+				&(si->dsk.dsk[i].avque) );
 
 			/*
 			** check if this line concerns the entire disk
@@ -726,9 +926,9 @@ photosyst(struct sstat *si)
 			*/
 			if (nr == 7)	/* full stats-line ? */
 			{
-				if ( !isrealdisk(diskname,
-				                 si->xdsk.xdsk[i].name,
-						 MAXDKNAM) )
+				if ( isdisk(0, 0, diskname,
+				                 &(si->dsk.dsk[i]),
+						 MAXDKNAM) != DSKTYPE)
 				       continue;
 			
 				if (++i >= MAXDSK-1)
@@ -736,8 +936,8 @@ photosyst(struct sstat *si)
 			}
 		}
 
-		si->xdsk.xdsk[i].name[0] = '\0'; /* set terminator for table */
-		si->xdsk.nrxdsk = i;
+		si->dsk.dsk[i].name[0] = '\0'; /* set terminator for table */
+		si->dsk.ndsk = i;
 
 		fclose(fp);
 
@@ -751,46 +951,75 @@ photosyst(struct sstat *si)
 	*/
 	if ( (fp = fopen("diskstats", "r")) != NULL)
 	{
-		char diskname[256];
-		i = 0;
+		char 		diskname[256];
+		struct perdsk	tmpdsk;
+
+		si->dsk.ndsk = 0;
+		si->dsk.nmdd = 0;
+		si->dsk.nlvm = 0;
 
 		while ( fgets(linebuf, sizeof(linebuf), fp) )
 		{
 			nr = sscanf(linebuf,
-			      "%*d %*d %255s %lld %*d %lld %*d "
+			      "%d %d %255s %lld %*d %lld %*d "
 			      "%lld %*d %lld %*d %*d %lld %lld",
-				diskname,
-				&(si->xdsk.xdsk[i].nread),
-				&(si->xdsk.xdsk[i].nrsect),
-				&(si->xdsk.xdsk[i].nwrite),
-				&(si->xdsk.xdsk[i].nwsect),
-				&(si->xdsk.xdsk[i].io_ms),
-				&(si->xdsk.xdsk[i].avque) );
+				&major, &minor, diskname,
+				&tmpdsk.nread,  &tmpdsk.nrsect,
+				&tmpdsk.nwrite, &tmpdsk.nwsect,
+				&tmpdsk.io_ms,  &tmpdsk.avque );
 
 			/*
 			** check if this line concerns the entire disk
 			** or just one of the partitions of a disk (to be
 			** skipped)
 			*/
-			if (nr == 7)	/* full stats-line ? */
+			if (nr == 9)	/* full stats-line ? */
 			{
-				if ( !isrealdisk(diskname,
-				                 si->xdsk.xdsk[i].name,
-						 MAXDKNAM) )
+				switch ( isdisk(major, minor, diskname,
+							 &tmpdsk, MAXDKNAM) )
+				{
+				   case NONTYPE:
 				       continue;
-			
-				if (++i >= MAXDSK-1)
+
+				   case DSKTYPE:
+					if (si->dsk.ndsk < MAXDSK-1)
+					  si->dsk.dsk[si->dsk.ndsk++] = tmpdsk;
 					break;
+
+				   case MDDTYPE:
+					if (si->dsk.nmdd < MAXMDD-1)
+					  si->dsk.mdd[si->dsk.nmdd++] = tmpdsk;
+					break;
+
+				   case LVMTYPE:
+					if (si->dsk.nlvm < MAXLVM-1)
+					  si->dsk.lvm[si->dsk.nlvm++] = tmpdsk;
+					break;
+				}
 			}
 		}
 
-		si->xdsk.xdsk[i].name[0] = '\0'; /* set terminator for table */
-		si->xdsk.nrxdsk = i;
+		/*
+ 		** set terminator for table
+ 		*/
+		si->dsk.dsk[si->dsk.ndsk].name[0] = '\0';
+		si->dsk.mdd[si->dsk.nmdd].name[0] = '\0';
+		si->dsk.lvm[si->dsk.nlvm].name[0] = '\0'; 
 
 		fclose(fp);
 	}
 
-	chdir(origdir);
+	/*
+ 	** get information about the shared memory statistics
+	*/
+	if ( shmctl(0, SHM_INFO, (struct shmid_ds *)&shminfo) != -1)
+	{
+		si->mem.shmrss = shminfo.shm_rss;
+		si->mem.shmswp = shminfo.shm_swp;
+	}
+
+	if ( chdir(origdir) == -1)
+		cleanstop(53);
 
 	/*
 	** fetch application-specific counters
@@ -803,17 +1032,19 @@ photosyst(struct sstat *si)
 
 /*
 ** set of subroutines to determine which disks should be monitored
-** and to translate long name strings into short name strings
+** and to translate name strings into (shorter) name strings
 */
 static void
-nullmodname(char *curname, char *newname, int maxlen)
+nullmodname(unsigned int major, unsigned int minor,
+		char *curname, struct perdsk *px, int maxlen)
 {
-	strncpy(newname, curname, maxlen-1);
-	*(newname+maxlen-1) = 0;
+	strncpy(px->name, curname, maxlen-1);
+	*(px->name+maxlen-1) = 0;
 }
 
 static void
-abbrevname1(char *curname, char *newname, int maxlen)
+abbrevname1(unsigned int major, unsigned int minor,
+		char *curname, struct perdsk *px, int maxlen)
 {
 	char	cutype[128];
 	int	hostnum, busnum, targetnum, lunnum;
@@ -821,34 +1052,151 @@ abbrevname1(char *curname, char *newname, int maxlen)
 	sscanf(curname, "%[^/]/host%d/bus%d/target%d/lun%d",
 			cutype, &hostnum, &busnum, &targetnum, &lunnum);
 
-	snprintf(newname, maxlen, "%c-h%db%dt%d", 
+	snprintf(px->name, maxlen, "%c-h%db%dt%d", 
 			cutype[0], hostnum, busnum, targetnum);
 }
 
 /*
-** table contains the names (in regexp format) of valid disks
-** to be supported, together a function to modify the name-strings
-** (i.e. to abbreviate long strings);
-** this table is used in the function isrealdisk()
+** recognize LVM logical volumes
+*/
+#define	NUMDMHASH	64
+#define	DMHASH(x,y)	(((x)+(y))%NUMDMHASH)	
+#define	MAPDIR		"/dev/mapper"
+
+struct devmap {
+	unsigned int	major;
+	unsigned int	minor;
+	char		name[MAXDKNAM];
+	struct devmap	*next;
+};
+
+static void
+lvmmapname(unsigned int major, unsigned int minor,
+		char *curname, struct perdsk *px, int maxlen)
+{
+	static int		firstcall = 1;
+	static struct devmap	*devmaps[NUMDMHASH], *dmp;
+	int			hashix;
+
+	/*
+ 	** setup a list of major-minor numbers of dm-devices with their
+	** corresponding name
+	*/
+	if (firstcall)
+	{
+		DIR		*dirp;
+		struct dirent	*dentry;
+		struct stat	statbuf;
+		char		path[64];
+
+		if ( (dirp = opendir(MAPDIR)) )
+		{
+			/*
+	 		** read every directory-entry and search for
+			** block devices
+			*/
+			while ( (dentry = readdir(dirp)) )
+			{
+				snprintf(path, sizeof path, "%s/%s", 
+						MAPDIR, dentry->d_name);
+
+				if ( stat(path, &statbuf) == -1 )
+					continue;
+
+				if ( ! S_ISBLK(statbuf.st_mode) )
+					continue;
+				/*
+ 				** allocate struct to store name
+				*/
+				if ( !(dmp = malloc(sizeof (struct devmap))))
+					continue;
+
+				/*
+ 				** store info in hash list
+				*/
+				strncpy(dmp->name, dentry->d_name, MAXDKNAM);
+				dmp->name[MAXDKNAM-1] = 0;
+				dmp->major 	= major(statbuf.st_rdev);
+				dmp->minor 	= minor(statbuf.st_rdev);
+
+				hashix = DMHASH(dmp->major, dmp->minor);
+
+				dmp->next	= devmaps[hashix];
+
+				devmaps[hashix]	= dmp;
+			}
+
+			closedir(dirp);
+		}
+
+		firstcall = 0;
+	}
+
+	/*
+ 	** find info in hash list
+	*/
+	hashix  = DMHASH(major, minor);
+	dmp	= devmaps[hashix];
+
+	while (dmp)
+	{
+		if (dmp->major == major && dmp->minor == minor)
+		{
+			/*
+		 	** info found in hash list; fill proper name
+			*/
+			strncpy(px->name, dmp->name, maxlen-1);
+			*(px->name+maxlen-1) = 0;
+			return;
+		}
+
+		dmp = dmp->next;
+	}
+
+	/*
+	** info not found in hash list; fill original name
+	*/
+	strncpy(px->name, curname, maxlen-1);
+	*(px->name+maxlen-1) = 0;
+}
+
+/*
+** this table is used in the function isdisk()
+**
+** table contains the names (in regexp format) of disks
+** to be recognized, together with a function to modify
+** the name-strings (i.e. to abbreviate long strings);
+** some frequently found names (like 'loop' and 'ram')
+** are also recognized to skip them as fast as possible
 */
 static struct {
 	char 	*regexp;
 	regex_t	compreg;
-	void	(*modname)(char *, char *, int);
+	void	(*modname)(unsigned int, unsigned int,
+				char *, struct perdsk *, int);
+	int	retval;
 } validdisk[] = {
-	{	"^sd[a-z][a-z]*$",			{0}, 	nullmodname, },
-	{	"^hd[a-z]$",				{0},	nullmodname, },
-	{	"^rd/c[0-9][0-9]*d[0-9][0-9]*$",	{0},	nullmodname, },
-	{	"^cciss/c[0-9][0-9]*d[0-9][0-9]*$",	{0},	nullmodname, },
-	{ 	"/host.*/bus.*/target.*/lun.*/disc",	{0},	abbrevname1, },
-
-	/* virtual disks */
-	{	"^xvd[a-z][a-z]*$",			{0},	nullmodname, },
-	{	"^dasd[a-z][a-z]*$",			{0},	nullmodname, },
+	{ "^ram[0-9][0-9]*$",			{0},  (void *)0,   NONTYPE, },
+	{ "^loop[0-9][0-9]*$",			{0},  (void *)0,   NONTYPE, },
+	{ "^sd[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
+	{ "^dm-[0-9][0-9]*$",			{0},  lvmmapname,  LVMTYPE, },
+	{ "^md[0-9][0-9]*$",			{0},  nullmodname, MDDTYPE, },
+	{ "^vd[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
+	{ "^hd[a-z]$",				{0},  nullmodname, DSKTYPE, },
+	{ "^rd/c[0-9][0-9]*d[0-9][0-9]*$",	{0},  nullmodname, DSKTYPE, },
+	{ "^cciss/c[0-9][0-9]*d[0-9][0-9]*$",	{0},  nullmodname, DSKTYPE, },
+	{ "^fio[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
+	{ "/host.*/bus.*/target.*/lun.*/disc",	{0},  abbrevname1, DSKTYPE, },
+	{ "^xvd[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
+	{ "^vd[a-z][a-z]*$",                    {0},  nullmodname, DSKTYPE, },
+	{ "^dasd[a-z][a-z]*$",			{0},  nullmodname, DSKTYPE, },
+	{ "^mmcblk[0-9][0-9]*$",		{0},  nullmodname, DSKTYPE, },
+	{ "^emcpower[a-z][a-z]*$",		{0},  nullmodname, DSKTYPE, },
 };
 
 static int
-isrealdisk(char *curname, char *newname, int maxlen)
+isdisk(unsigned int major, unsigned int minor,
+           char *curname, struct perdsk *px, int maxlen)
 {
 	static int	firstcall = 1;
 	register int	i;
@@ -871,34 +1219,35 @@ isrealdisk(char *curname, char *newname, int maxlen)
 			/*
 			** name-string recognized; modify name-string
 			*/
-			(*validdisk[i].modname)(curname, newname, maxlen);
+			if (validdisk[i].retval != NONTYPE)
+				(*validdisk[i].modname)(major, minor,
+						curname, px, maxlen);
 
-			return 1;
+			return validdisk[i].retval;
 		}
 	}
 
-	return 0;
+	return NONTYPE;
 }
 
 /*
 ** LINUX SPECIFIC:
-** Determine boot-time of this system (as number of seconds since 1-1-1970).
+** Determine boot-time of this system (as number of jiffies since 1-1-1970).
 */
-time_t	getbootlinux(long);
-
-time_t
+unsigned long long
 getbootlinux(long hertz)
 {
-	int     	cpid;
-	char  	  	tmpbuf[1280];
-	FILE    	*fp;
-	unsigned long 	startticks;
-	time_t		boottime = 0;
+	int    		 	cpid;
+	char  	  		tmpbuf[1280];
+	FILE    		*fp;
+	unsigned long 		startticks;
+	unsigned long long	bootjiffies = 0;
+	struct timespec		ts;
 
 	/*
 	** dirty hack to get the boottime, since the
 	** Linux 2.6 kernel (2.6.5) does not return a proper
-	** boottime-value with the times() system call :-(
+	** boottime-value with the times() system call   :-(
 	*/
 	if ( (cpid = fork()) == 0 )
 	{
@@ -910,10 +1259,14 @@ getbootlinux(long hertz)
 	else
 	{
 		/*
-		** parent determines start-time (in clock-ticks since boot) 
-		** of the child and calculates the boottime in seconds
+		** parent determines start-time (in jiffies since boot) 
+		** of the child and calculates the boottime in jiffies
 		** since 1-1-1970
 		*/
+		(void) clock_gettime(CLOCK_REALTIME, &ts);	// get current
+		bootjiffies = 1LL * ts.tv_sec  * hertz +
+		              1LL * ts.tv_nsec * hertz / 1000000000LL;
+
 		snprintf(tmpbuf, sizeof tmpbuf, "/proc/%d/stat", cpid);
 
 		if ( (fp = fopen(tmpbuf, "r")) != NULL)
@@ -923,7 +1276,7 @@ getbootlinux(long hertz)
 			                "%*d %*d %*d %*d %*d %*d %lu",
 			                &startticks) == 1)
 			{
-				boottime = time(0) - startticks / hertz;
+				bootjiffies -= startticks;
 			}
 
 			fclose(fp);
@@ -936,7 +1289,7 @@ getbootlinux(long hertz)
 		(void) wait((int *)0);
 	}
 
-	return boottime;
+	return bootjiffies;
 }
 
 #if	HTTPSTATS
