@@ -4,11 +4,11 @@
 ** The program 'atop' offers the possibility to view the activity of 
 ** the system on system-level as well as process-level.
 ** ==========================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        September 2002
 ** --------------------------------------------------------------------------
-** Copyright (C) 2000-2005 Gerlof Langeveld
+** Copyright (C) 2000-2010 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -26,6 +26,41 @@
 ** --------------------------------------------------------------------------
 **
 ** $Log: rawlog.c,v $
+** Revision 1.32  2010/11/26 06:06:35  gerlof
+** Cosmetic.
+**
+** Revision 1.31  2010/11/17 12:43:31  gerlof
+** The flag -r followed by exactly 8 'y' characters is not considered
+** as 8 days ago, but as a literal filename.
+**
+** Revision 1.30  2010/10/23 14:03:03  gerlof
+** Counters for total number of running and sleep threads (JC van Winkel).
+**
+** Revision 1.29  2010/04/23 13:55:52  gerlof
+** Get rid of all setuid-root privs before activating other program.
+**
+** Revision 1.28  2010/04/23 12:19:35  gerlof
+** Modified mail-address in header.
+**
+** Revision 1.27  2010/04/16 12:55:16  gerlof
+** Automatically start another version of atop if the logfile to
+** be read has not been created by the current version.
+**
+** Revision 1.26  2010/03/02 13:55:29  gerlof
+** Struct stat size did not fit in short any more (modified to int).
+**
+** Revision 1.25  2009/12/17 08:16:06  gerlof
+** Introduce branch-key to go to specific time in raw file.
+**
+** Revision 1.24  2009/11/27 15:26:29  gerlof
+** Added possibility to specify y[y..] als filename for -r flag
+** to access file of yesterday, day before yesterday, etc.
+**
+** Revision 1.23  2009/11/27 14:28:14  gerlof
+** Rollback a "transaction" when not all parts could be
+** written to the logfile (e.g. file system full) to avoid
+** a corrupted logfile.
+**
 ** Revision 1.22  2008/01/07 10:18:05  gerlof
 ** Implement possibility to make summaries.
 **
@@ -76,7 +111,7 @@
 **
 ** Revision 1.7  2003/01/17 07:33:39  gerlof
 ** Modified process statistics: add command-line.
-** Implement compatibility for old pstat-structure read from logfiles.
+** Implement compatibility for old tstat-structure read from logfiles.
 **
 ** Revision 1.6  2002/10/30 13:46:11  gerlof
 ** Generate notification for statistics since boot.
@@ -100,7 +135,7 @@
 **
 */
 
-static const char rcsid[] = "$Id: rawlog.c,v 1.22 2008/01/07 10:18:05 gerlof Exp $";
+static const char rcsid[] = "$Id: rawlog.c,v 1.32 2010/11/26 06:06:35 gerlof Exp $";
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -110,15 +145,19 @@ static const char rcsid[] = "$Id: rawlog.c,v 1.22 2008/01/07 10:18:05 gerlof Exp
 #include <stdio.h>
 #include <errno.h>
 #include <fcntl.h>
-#include <unistd.h>
 #include <stdlib.h>
 #include <signal.h>
 #include <ctype.h>
 #include <sys/utsname.h>
 #include <string.h>
+#include <regex.h>
 #include <zlib.h>
+#include <sys/time.h>
+#include <sys/resource.h>
+#include <unistd.h>
 
 #include "atop.h"
+#include "showgeneric.h"
 #include "photoproc.h"
 #include "photosyst.h"
 
@@ -145,12 +184,14 @@ struct rawheader {
 	unsigned int	magic;
 
 	unsigned short	aversion;	/* creator atop version with MSB */
-	unsigned short	sstatlen;	/* length of struct sstat        */
-	unsigned short	pstatlen;	/* length of struct pstat        */
+	unsigned short	future1;	/* can be reused 		 */
+	unsigned short	future2;	/* can be reused 		 */
 	unsigned short	rawheadlen;	/* length of struct rawheader    */
 	unsigned short	rawreclen;	/* length of struct rawrecord    */
 	unsigned short	hertz;		/* clock interrupts per second   */
-	unsigned short	sfuture[5];	/* future use                    */
+	unsigned short	sfuture[6];	/* future use                    */
+	unsigned int	sstatlen;	/* length of struct sstat        */
+	unsigned int	tstatlen;	/* length of struct tstat        */
 	struct utsname	utsname;	/* info about this system        */
 	char		cfuture[8];	/* future use                    */
 
@@ -169,21 +210,28 @@ struct rawrecord {
 	unsigned short	sfuture[3];	/* future use                   */
 
 	unsigned int	scomplen;	/* length of compressed sstat   */
-	unsigned int	pcomplen;	/* length of compressed pstat's */
+	unsigned int	pcomplen;	/* length of compressed tstat's */
 	unsigned int	interval;	/* interval (number of seconds) */
-	unsigned int	nlist;		/* number of processes in list  */
-	unsigned int	npresent;	/* total number of processes    */
+	unsigned int	ndeviat;	/* number of tasks in list      */
+	unsigned int	nactproc;	/* number of processes in list  */
+	unsigned int	ntask;		/* total number of tasks        */
+	unsigned int    totproc;	/* total number of processes	*/
+	unsigned int    totrun;		/* number of running  threads	*/
+	unsigned int    totslpi;	/* number of sleeping threads(S)*/
+	unsigned int    totslpu;	/* number of sleeping threads(D)*/
+	unsigned int	totzomb;	/* number of zombie processes   */
 	unsigned int	nexit;		/* number of exited processes   */
-	unsigned int	nzombie;	/* number of zombie processes   */
+	unsigned int	noverflow;	/* number of overflow processes */
 	unsigned int	ifuture[6];	/* future use                   */
 };
 
 static int	getrawrec  (int, struct rawrecord *, int);
 static int	getrawsstat(int, struct sstat *, int);
-static int	getrawpstat(int, struct pstat *, int, int);
+static int	getrawtstat(int, struct tstat *, int, int);
 static int	rawwopen(void);
 static int	lookslikedatetome(char *);
 static void	testcompval(int, char *);
+static void	try_other_version(int, int);
 
 /*
 ** write a raw record to file
@@ -191,17 +239,19 @@ static void	testcompval(int, char *);
 */
 char
 rawwrite(time_t curtime, int numsecs, 
-	 struct sstat *ss, struct pstat *ps,
-	 int nlist, int npresent, int nzombie,
-	 int nexit, char flag)
+	 struct sstat *ss, struct tstat *ts, struct tstat **proclist,
+	 int ndeviat, int ntask, int nactproc,
+         int totproc, int totrun, int totslpi, int totslpu, int totzomb, 
+         int nexit, unsigned int noverflow, char flag)
 {
 	static int		rawfd = -1;
 	struct rawrecord	rr;
 	int			rv;
+	struct stat		filestat;
 
 	Byte			scompbuf[sizeof(struct sstat)], *pcompbuf;
 	unsigned long		scomplen = sizeof scompbuf;
-	unsigned long		pcomplen = sizeof(struct pstat) * nlist;
+	unsigned long		pcomplen = sizeof(struct tstat) * ndeviat;
 
 	/*
 	** first call:
@@ -210,6 +260,13 @@ rawwrite(time_t curtime, int numsecs,
 	if (rawfd == -1)
 		rawfd = rawwopen();
 
+	/*
+ 	** register current size of file in order to "roll back"
+	** writes that have been done while not *all* writes could
+	** succeed, e.g. when file system full
+	*/
+	(void) fstat(rawfd, &filestat);
+
 	/*
 	** compress system- and process-level statistics
 	*/
@@ -218,13 +275,11 @@ rawwrite(time_t curtime, int numsecs,
 
 	testcompval(rv, "compress");
 
-	if ( (pcompbuf = malloc(pcomplen)) == NULL)
-	{
-		perror("atop malloc");
-		cleanstop(7);
-	}
+	pcompbuf = malloc(pcomplen);
 
-	rv = compress(pcompbuf, &pcomplen, (Byte *)ps, (unsigned long)pcomplen);
+	ptrverify(pcompbuf, "Malloc failed for compression buffer\n");
+
+	rv = compress(pcompbuf, &pcomplen, (Byte *)ts, (unsigned long)pcomplen);
 
 	testcompval(rv, "compress");
 
@@ -236,20 +291,34 @@ rawwrite(time_t curtime, int numsecs,
 	rr.curtime	= curtime;
 	rr.interval	= numsecs;
 	rr.flags	= 0;
-	rr.nlist	= nlist;
-	rr.npresent	= npresent;
-	rr.nzombie	= nzombie;
+	rr.ndeviat	= ndeviat;
+	rr.nactproc	= nactproc;
+	rr.ntask	= ntask;
 	rr.nexit	= nexit;
+	rr.noverflow	= noverflow;
+	rr.totproc	= totproc;
+	rr.totrun	= totrun;
+	rr.totslpi	= totslpi;
+	rr.totslpu	= totslpu;
+	rr.totzomb	= totzomb;
 	rr.scomplen	= scomplen;
 	rr.pcomplen	= pcomplen;
 
 	if (flag&RRBOOT)
 		rr.flags |= RRBOOT;
 
+	if (supportflags & NETATOP)
+		rr.flags |= RRNETATOP;
+
+	if (supportflags & NETATOPD)
+		rr.flags |= RRNETATOPD;
+
 	if ( write(rawfd, &rr, sizeof rr) == -1)
 	{
 		fprintf(stderr, "%s - ", rawname);
 		perror("write raw record");
+		if ( ftruncate(rawfd, filestat.st_size) == -1)
+			cleanstop(8);
 		cleanstop(7);
 	}
 
@@ -260,6 +329,8 @@ rawwrite(time_t curtime, int numsecs,
 	{
 		fprintf(stderr, "%s - ", rawname);
 		perror("write raw status record");
+		if ( ftruncate(rawfd, filestat.st_size) == -1)
+			cleanstop(8);
 		cleanstop(7);
 	}
 
@@ -270,6 +341,8 @@ rawwrite(time_t curtime, int numsecs,
 	{
 		fprintf(stderr, "%s - ", rawname);
 		perror("write raw process record");
+		if ( ftruncate(rawfd, filestat.st_size) == -1)
+			cleanstop(8);
 		cleanstop(7);
 	}
 
@@ -322,16 +395,17 @@ rawwopen()
 		}
 
 		if ( rh.sstatlen	!= sizeof(struct sstat)		||
-		     rh.pstatlen	!= sizeof(struct pstat)		||
+		     rh.tstatlen	!= sizeof(struct tstat)		||
 	    	     rh.rawheadlen	!= sizeof(struct rawheader)	||
 		     rh.rawreclen	!= sizeof(struct rawrecord)	||
-		     rh.supportflags	!= supportflags			  )
+		     rh.supportflags	!= (supportflags & ~(NETATOP|NETATOPD)))
 		{
 			fprintf(stderr,
-				"%s exists but has incompatible header\n",
+				"existing file %s has incompatible header\n",
 				rawname);
 
-			if (rh.aversion & 0x8000)
+			if (rh.aversion & 0x8000 &&
+			   (rh.aversion & 0x7fff) != getnumvers())
 			{
 				fprintf(stderr,
 					"(created by version %d.%d - "
@@ -365,10 +439,10 @@ rawwopen()
 	rh.magic	= MYMAGIC;
 	rh.aversion	= getnumvers() | 0x8000;
 	rh.sstatlen	= sizeof(struct sstat);
-	rh.pstatlen	= sizeof(struct pstat);
+	rh.tstatlen	= sizeof(struct tstat);
 	rh.rawheadlen	= sizeof(struct rawheader);
 	rh.rawreclen	= sizeof(struct rawrecord);
-	rh.supportflags	= supportflags;
+	rh.supportflags	= (supportflags & ~(NETATOP|NETATOPD));
 	rh.osrel	= osrel;
 	rh.osvers	= osvers;
 	rh.ossub	= ossub;
@@ -393,13 +467,15 @@ rawwopen()
 #define	OFFCHUNK	256
 
 void
-rawread(unsigned int begintime, unsigned int endtime)
+rawread(void)
 {
-	int			rawfd;
+	int			i, j, rawfd, len;
+	char			*py;
 	struct rawheader	rh;
 	struct rawrecord	rr;
 	struct sstat		devsstat;
-	struct pstat		*devpstat;
+	struct tstat		*devtstat;
+	struct tstat		**devpstat;
 
 	struct stat		filestat;
 
@@ -415,7 +491,7 @@ rawread(unsigned int begintime, unsigned int endtime)
 	time_t			timenow;
 	struct tm		*tp;
 
-	switch ( strlen(rawname) )
+	switch ( len = strlen(rawname) )
 	{
 	   /*
 	   ** if no filename is specified, assemble the name of the raw file
@@ -449,8 +525,43 @@ rawread(unsigned int begintime, unsigned int endtime)
 			snprintf(rawname, RAWNAMESZ, "%s/atop_%s",
 				BASEPATH, 
 				savedname);
+			break;
 		}
-		break;
+
+	   /*
+	   ** if one or more 'y' (yesterday) characters are used and that
+	   ** string is not known as an existing file, the standard logfile
+	   ** is shown from N days ago (N is determined by the number
+	   ** of y's).
+	   */
+	   default:
+		if ( access(rawname, F_OK) == 0) 
+			break;		/* existing file */
+
+		/*
+		** make a string existing of y's to compare with
+		*/
+		py = malloc(len+1);
+
+		ptrverify(py, "Malloc failed for 'yes' sequence\n");
+
+		memset(py, 'y', len);
+		*(py+len) = '\0';
+
+		if ( strcmp(rawname, py) == 0 )
+		{
+			timenow	 = time(0);
+			timenow -= len*3600*24;
+			tp	 = localtime(&timenow);
+
+			snprintf(rawname, RAWNAMESZ, "%s/atop_%04d%02d%02d",
+				BASEPATH, 
+				tp->tm_year+1900,
+				tp->tm_mon+1,
+				tp->tm_mday);
+		}
+
+		free(py);
 	}
 
 	/*
@@ -476,20 +587,25 @@ rawread(unsigned int begintime, unsigned int endtime)
 		** compressed raw file to be decompressed via gunzip
 		*/
 		fprintf(stderr, "Decompressing logfile ....\n");
-
-		snprintf(tmpname2, sizeof tmpname2, "/tmp/atopwrk%d", getpid());
-		snprintf(command,  sizeof command, "gunzip -c %s > %s",
-							tmpname1, tmpname2);
-		system (command);
-
-		if ( (rawfd = open(tmpname2, O_RDONLY)) == -1)
+		snprintf(tmpname2, sizeof tmpname2, "/tmp/atopwrkXXXXXX");
+		rawfd = mkstemp(tmpname2);
+		if (rawfd == -1)
 		{
 			fprintf(stderr, "%s - ", rawname);
-			perror("open decompressed raw file");
+			perror("creating decompression temp file");
 			cleanstop(7);
 		}
 
+		snprintf(command,  sizeof command, "gunzip -c %s > %s",
+							tmpname1, tmpname2);
+		const int system_res = system (command);
 		unlink(tmpname2);
+
+		if (system_res)
+		{
+			fprintf(stderr, "%s - gunzip failed", rawname);
+			cleanstop(7);
+		}
 	}
 
 	/*
@@ -512,23 +628,39 @@ rawread(unsigned int begintime, unsigned int endtime)
 	** magic okay, but file-layout might have been modified
 	*/
 	if (rh.sstatlen   != sizeof(struct sstat)		||
-	    rh.pstatlen   != sizeof(struct pstat)		||
+	    rh.tstatlen   != sizeof(struct tstat)		||
 	    rh.rawheadlen != sizeof(struct rawheader)		||
 	    rh.rawreclen  != sizeof(struct rawrecord)		  )
 	{
 		fprintf(stderr,
 			"raw file %s has incompatible format\n", rawname);
 
-		if (rh.aversion & 0x8000)
+		if (rh.aversion & 0x8000 &&
+       		   (rh.aversion & 0x7fff) != getnumvers())
 		{
 			fprintf(stderr,
 				"(created by version %d.%d - "
 				"current version %d.%d)\n",
 				(rh.aversion >> 8) & 0x7f,
-				 rh.aversion & 0xff,
+				 rh.aversion       & 0xff,
 				 getnumvers() >> 8,
 				 getnumvers() & 0x7f);
 		}
+		else
+		{
+			fprintf(stderr,
+				"(files from other system architectures might"
+				" be binary incompatible)\n");
+		}
+
+		close(rawfd);
+
+		if (((rh.aversion >> 8) & 0x7f) != (getnumvers()   >> 8) ||
+		     (rh.aversion       & 0xff) != (getnumvers() & 0x7f)   )
+		{
+			try_other_version((rh.aversion >> 8) & 0x7f,
+			                rh.aversion       & 0xff);
+		}
 
 		cleanstop(7);
 	}
@@ -549,11 +681,9 @@ rawread(unsigned int begintime, unsigned int endtime)
 	/*
 	** allocate a list for backtracking of rawrecord-offsets
 	*/
-	if ( (offlist = malloc(sizeof(off_t) * OFFCHUNK)) == NULL)
-	{
-		perror("atop/atopsar malloc");
-		cleanstop(7);
-	}
+	offlist = malloc(sizeof(off_t) * OFFCHUNK);
+
+	ptrverify(offlist, "Malloc failed for backtrack list\n");
 
 	offsize = OFFCHUNK;
 
@@ -573,19 +703,18 @@ rawread(unsigned int begintime, unsigned int endtime)
 
 			/*
 			** store the offset of the raw record in the offset list
-			** if case of offset list overflow, extend the list
+			** in case of offset list overflow, extend the list
 			*/
 			*(offlist+offcur) = lseek(rawfd, 0, SEEK_CUR) -
 								rh.rawreclen;
 
 			if ( ++offcur >= offsize )
 			{
-				if ( (offlist = realloc(offlist,
-				     (offsize+OFFCHUNK)*sizeof(off_t))) ==NULL)
-				{
-					perror("atop/atopsar realloc");
-					cleanstop(7);
-				}
+				offlist = realloc(offlist,
+				             (offsize+OFFCHUNK)*sizeof(off_t));
+
+				ptrverify(offlist,
+				        "Realloc failed for backtrack list\n");
 
 				offsize+= OFFCHUNK;
 			}
@@ -594,14 +723,22 @@ rawread(unsigned int begintime, unsigned int endtime)
 			** check if this sample is within the time-range
 			** specified with the -b and -e flags (if any)
 			*/
-			if ( (begintime && begintime > secsinday) ||
-			     (endtime   && endtime   < secsinday)   )
+			if ( (begintime && begintime > secsinday) )
 			{
-				lastcmd = '\0';
+				lastcmd = 1;
 				lseek(rawfd, rr.scomplen+rr.pcomplen, SEEK_CUR);
 				continue;
 			}
-	
+
+			begintime = 0;	// allow earlier times from now on
+
+			if ( (endtime && endtime < secsinday) )
+			{
+				free(offlist);
+				close(rawfd);
+				return;
+			}
+
 			/*
 			** allocate space, read compressed system-level
 			** statistics and decompress
@@ -613,15 +750,20 @@ rawread(unsigned int begintime, unsigned int endtime)
 			** allocate space, read compressed process-level
 			** statistics and decompress
 			*/
-			if ( (devpstat =
-			       malloc(sizeof(struct pstat) * rr.nlist)) ==NULL)
-			{
-				perror("atop/atopsar malloc");
-				cleanstop(7);
-			}
+			devtstat = malloc(sizeof(struct tstat) * rr.ndeviat);
 
-			if ( !getrawpstat(rawfd, devpstat,
-					rr.pcomplen, rr.nlist) )
+			ptrverify(devtstat,
+			          "Malloc failed for %d stored tasks\n",
+			          rr.ndeviat);
+
+			devpstat = malloc(sizeof(struct tstat *) * rr.nactproc);
+
+			ptrverify(devpstat,
+			          "Malloc failed for %d stored processes\n",
+			          rr.nactproc);
+
+			if ( !getrawtstat(rawfd, devtstat,
+						rr.pcomplen, rr.ndeviat) )
 				cleanstop(7);
 
 			/*
@@ -630,6 +772,16 @@ rawread(unsigned int begintime, unsigned int endtime)
 			*/
 			sampcnt++;
 
+			if (rr.flags & RRNETATOP)
+				supportflags |=  NETATOP;
+			else
+				supportflags &= ~NETATOP;
+
+			if (rr.flags & RRNETATOPD)
+				supportflags |=  NETATOPD;
+			else
+				supportflags &= ~NETATOPD;
+
 			flags = rr.flags & RRBOOT;
 
 			(void) fstat(rawfd, &filestat);
@@ -638,16 +790,25 @@ rawread(unsigned int begintime, unsigned int endtime)
 			     lseek(rawfd, (off_t)0, SEEK_CUR) <= rh.rawreclen)
 				flags |= RRLAST;
 
+			for (i=j=0; i < rr.ndeviat; i++)
+			{
+				if ( (devtstat+i)->gen.isproc)
+					devpstat[j++] = devtstat+i;
+			}
+
 			lastcmd = (vis.show_samp)(rr.curtime, rr.interval,
-						&devsstat,    devpstat,
-					 	rr.nlist,     rr.npresent,
-						rr.nzombie,   rr.nexit,
-						flags);
+			             &devsstat, devtstat, devpstat,
+		 	             rr.ndeviat,  rr.ntask, rr.nactproc,
+			             rr.totproc, rr.totrun, rr.totslpi,
+			             rr.totslpu, rr.totzomb, rr.nexit,
+			             rr.noverflow, flags);
+
+			free(devtstat);
 			free(devpstat);
 	
 			switch (lastcmd)
 			{
-			   case 'T':
+			   case MSAMPPREV:
 				if (offcur >= 2)
 					offcur-= 2;
 				else
@@ -656,12 +817,22 @@ rawread(unsigned int begintime, unsigned int endtime)
 				lseek(rawfd, *(offlist+offcur), SEEK_SET);
 				break;
 
-			   case 'r':
+			   case MRESET:
 				lseek(rawfd, *offlist, SEEK_SET);
 				offcur = 1;
+				break;
+
+			   case MSAMPBRANCH:
+				if (begintime && begintime < secsinday)
+				{
+					lseek(rawfd, *offlist, SEEK_SET);
+					offcur = 1;
+				}
 			}
 		}
 
+		begintime = 0;	// allow earlier times from now on
+
 		if (offcur >= 1)
 			offcur--;
 
@@ -693,10 +864,8 @@ getrawsstat(int rawfd, struct sstat *sp, int complen)
 	int		rv;
 
 	if ( (compbuf = malloc(complen)) == NULL)
-	{
-		perror("atop/atopsar malloc");
-		cleanstop(7);
-	}
+
+	ptrverify(compbuf, "Malloc failed for reading compressed sysstats\n");
 
 	if ( read(rawfd, compbuf, complen) < complen)
 		return 0;
@@ -714,17 +883,15 @@ getrawsstat(int rawfd, struct sstat *sp, int complen)
 ** read the process-level statistics from the current offset
 */
 static int
-getrawpstat(int rawfd, struct pstat *pp, int complen, int nlist)
+getrawtstat(int rawfd, struct tstat *pp, int complen, int ndeviat)
 {
 	Byte		*compbuf;
-	unsigned long	uncomplen = sizeof(struct pstat) * nlist;
+	unsigned long	uncomplen = sizeof(struct tstat) * ndeviat;
 	int		rv;
 
-	if ( (compbuf = malloc(complen)) == NULL)
-	{
-		perror("atop/atopsar malloc");
-		cleanstop(7);
-	}
+	compbuf = malloc(complen);
+
+	ptrverify(compbuf, "Malloc failed for reading compressed procstats\n");
 
 	if ( read(rawfd, compbuf, complen) < complen)
 		return 0;
@@ -793,3 +960,59 @@ testcompval(int rv, char *func)
 		cleanstop(7);
 	}
 }
+
+/*
+** try to activate another atop- or atopsar-version
+** to read this logfile
+*/
+static void
+try_other_version(int majorversion, int minorversion)
+{
+	char		tmpbuf[1024], *p;
+	extern char	**argvp;
+	int		fds;
+	struct rlimit	rlimit;
+	int 		setresuid(uid_t, uid_t, uid_t);
+
+	/*
+ 	** prepare name of executable file
+	** the current pathname (if any) is stripped off
+	*/
+	snprintf(tmpbuf, sizeof tmpbuf, "%s-%d.%d",
+		(p = strrchr(*argvp, '/')) ? p+1 : *argvp,
+			majorversion, minorversion);
+
+	fprintf(stderr, "trying to activate %s....\n", tmpbuf);
+
+	/*
+	** be sure no open file descriptors are passed
+	** except stdin, stdout en stderr
+	*/
+	(void) getrlimit(RLIMIT_NOFILE, &rlimit);
+
+	for (fds=3; fds < rlimit.rlim_cur; fds++)
+		close(fds);
+
+	/*
+	** be absolutely sure not to pass setuid-root privileges
+	** to the loaded program; errno EAGAIN and ENOMEM are not
+	** acceptable!
+	*/
+	if ( setresuid(getuid(), getuid(), getuid()) == -1 && errno != EPERM)
+	{
+		fprintf(stderr, "not possible to drop root-privileges!\n");
+		exit(1);
+	}
+
+	/*
+ 	** load alternative executable image
+	** at this moment the saved-uid might still be set
+	** to 'root' but this is reset at the moment of exec
+	*/
+	(void) execvp(tmpbuf, argvp);
+
+	/*
+	** point of no return, except when exec failed
+	*/
+	fprintf(stderr, "activation of %s failed!\n", tmpbuf);
+}
