@@ -7,8 +7,8 @@
 ** Include-file describing prototypes and structures for visualization
 ** of counters.
 ** ================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        July 2002
 **
 ** This program is free software; you can redistribute it and/or modify it
@@ -32,13 +32,27 @@ struct syscap {
 	count_t	availnet;
 };
 
-struct selection {
+struct pselection {
 	char	username[256];
 	uid_t	userid[MAXUSERSEL];
 
-	char	procname[64];
-	int	procnamesz;
-	regex_t	procregex;
+	char	progname[64];
+	int	prognamesz;
+	regex_t	progregex;
+};
+
+struct sselection {
+	char	lvmname[64];	// logical volume selection
+	int	lvmnamesz;
+	regex_t	lvmregex;
+
+	char	dskname[64];	// disk selection
+	int	dsknamesz;
+	regex_t	dskregex;
+
+	char	itfname[64];	// network interface selection
+	int	itfnamesz;
+	regex_t	itfregex;
 };
 
 /*
@@ -47,6 +61,7 @@ struct selection {
 #define	COLORLOW	2
 #define	COLORMED	3
 #define	COLORHIGH	4
+#define	COLORTHR	5
 
 /*
 ** list with keystrokes/flags
@@ -58,6 +73,7 @@ struct selection {
 #define	MPROCSCH	's'
 #define	MPROCVAR	'v'
 #define	MPROCARG	'c'
+#define MPROCOWN	'o'
 
 #define	MCUMUSER	'u'
 #define	MCUMPROC	'p'
@@ -68,23 +84,28 @@ struct selection {
 #define	MSORTNET	'N'
 #define	MSORTAUTO	'A'
 
+#define	MTHREAD		'y'
 #define	MCOLORS		'x'
 #define	MSYSFIXED	'f'
+#define	MSYSNOSORT	'F'
 #define	MSYSLIMIT	'l'
 
-#define	MSELPROC	'P'
 #define	MSELUSER	'U'
+#define	MSELPROC	'P'
+#define	MSELSYS		'S'
 
 #define	MALLPROC	'a'
 #define	MKILLPROC	'k'
 #define	MLISTFW		0x06
 #define	MLISTBW		0x02
+#define MREDRAW         0x0c
 #define	MINTERVAL	'i'
 #define	MPAUSE		'z'
 #define	MQUIT		'q'
 #define	MRESET		'r'
 #define	MSAMPNEXT	't'
 #define	MSAMPPREV	'T'
+#define	MSAMPBRANCH	'b'
 #define	MVERSION	'V'
 #define	MAVGVAL		'1'
 #define	MHELP1		'?'
@@ -93,21 +114,22 @@ struct selection {
 /*
 ** general function prototypes
 */
-void	totalcap   (struct syscap *, struct sstat *, struct pstat *, int);
-void	pricumproc (struct pstat *, int, int, int, int, int, int);
+void	totalcap   (struct syscap *, struct sstat *, struct tstat **, int);
+void	pricumproc (struct sstat *,  struct tstat **, int, int, int,
+			int, int, int, int, int, unsigned int, int, int);
 
-void	showgenproc(struct pstat *, double, int, int);
-void	showmemproc(struct pstat *, double, int, int);
-void	showdskproc(struct pstat *, double, int, int);
-void	shownetproc(struct pstat *, double, int, int);
-void	showvarproc(struct pstat *, double, int, int);
-void	showschproc(struct pstat *, double, int, int);
-void	showtotproc(struct pstat *, double, int, int);
-void	showcmdproc(struct pstat *, double, int, int);
+void	showgenproc(struct tstat *, double, int, int);
+void	showmemproc(struct tstat *, double, int, int);
+void	showdskproc(struct tstat *, double, int, int);
+void	shownetproc(struct tstat *, double, int, int);
+void	showvarproc(struct tstat *, double, int, int);
+void	showschproc(struct tstat *, double, int, int);
+void	showtotproc(struct tstat *, double, int, int);
+void	showcmdproc(struct tstat *, double, int, int);
 
 void	printg     (const char *, ...);
-int	prisyst(struct sstat  *, int, int, int, int, int, char *,
-                int, int, int);
-int	priproc(struct pstat  *, int, int, int, int, int, char, char,
-	        struct syscap *, struct selection *, int, int);
-void	priphead(int, int, char, char, char);
+int	prisyst(struct sstat  *, int, int, int, int, struct sselection *,
+			char *, int, int, int, int, int);
+int	priproc(struct tstat  **, int, int, int, int, int, char, char,
+	        struct syscap *, int, int);
+void	priphead(int, int, char *, char *, char);
