@@ -7,12 +7,12 @@
 ** This source-file contains various functions to a.o. format the
 ** time-of-day, the cpu-time consumption and the memory-occupation. 
 ** ==========================================================================
-** Author:      Gerlof Langeveld - AT Computing, Nijmegen, Holland
-** E-mail:      gerlof@ATComputing.nl
+** Author:      Gerlof Langeveld
+** E-mail:      gerlof.langeveld@atoptool.nl
 ** Date:        November 1996
 ** LINUX-port:  June 2000
 ** --------------------------------------------------------------------------
-** Copyright (C) 2000-2005 Gerlof Langeveld
+** Copyright (C) 2000-2010 Gerlof Langeveld
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -30,6 +30,28 @@
 ** --------------------------------------------------------------------------
 **
 ** $Log: various.c,v $
+** Revision 1.21  2010/11/12 06:16:16  gerlof
+** Show all parts of timestamp in header line, even when zero.
+**
+** Revision 1.20  2010/05/18 19:21:08  gerlof
+** Introduce CPU frequency and scaling (JC van Winkel).
+**
+** Revision 1.19  2010/04/28 18:21:11  gerlof
+** Cast value larger than 4GB to long long.
+**
+** Revision 1.18  2010/04/23 12:19:35  gerlof
+** Modified mail-address in header.
+**
+** Revision 1.17  2010/03/26 11:52:45  gerlof
+** Introduced unit of Tbytes for memory-usage.
+**
+** Revision 1.16  2009/12/17 08:28:38  gerlof
+** Express CPU-time usage in days and hours for large values.
+**
+** Revision 1.15  2009/12/10 08:50:39  gerlof
+** Introduction of a new function to convert number of seconds
+** to a string indicating days, hours, minutes and seconds.
+**
 ** Revision 1.14  2007/02/13 10:32:47  gerlof
 ** Removal of external declarations.
 ** Removal of function getpagesz().
@@ -76,7 +98,7 @@
 **
 */
 
-static const char rcsid[] = "$Id: various.c,v 1.14 2007/02/13 10:32:47 gerlof Exp $";
+static const char rcsid[] = "$Id: various.c,v 1.21 2010/11/12 06:16:16 gerlof Exp $";
 
 #include <sys/types.h>
 #include <sys/param.h>
@@ -89,6 +111,8 @@ static const char rcsid[] = "$Id: various.c,v 1.14 2007/02/13 10:32:47 gerlof Ex
 #include <unistd.h>
 #include <ctype.h>
 #include <stdlib.h>
+#include <errno.h>
+#include <stdarg.h>
 
 #include "atop.h"
 #include "acctproc.h"
@@ -241,31 +265,78 @@ val2valstr(count_t value, char *strvalue, int width, int avg, int nsecs)
 	return strvalue;
 }
 
+#define DAYSECS 	(24*60*60)
+#define HOURSECS	(60*60)
+#define MINSECS 	(60)
+
+/*
+** Function val2elapstr() converts a value (number of seconds)
+** to an ascii-string of up to max 13 positions in NNNdNNhNNmNNs
+** stored in strvalue (at least 14 positions).
+** returnvalue: number of bytes stored
+*/
+int
+val2elapstr(int value, char *strvalue)
+{
+        char	*p=strvalue, doshow=0;
+
+        if (value > DAYSECS) 
+        {
+                p+=sprintf(p, "%dd", value/DAYSECS);
+                value %= DAYSECS;
+		doshow = 1;
+        }
+
+        if (value > HOURSECS || doshow) 
+        {
+                p+=sprintf(p, "%dh", value/HOURSECS);
+                value %= HOURSECS;
+		doshow = 1;
+        }
+
+        if (value > MINSECS || doshow) 
+        {
+                p+=sprintf(p, "%dm", value/MINSECS);
+                value %= MINSECS;
+		doshow = 1;
+        }
+
+        if (value || doshow) 
+        {
+                p+=sprintf(p, "%ds", value);
+		doshow = 1;
+        }
+
+        return p-strvalue;
+}
+
+
 /*
 ** Function val2cpustr() converts a value (number of milliseconds)
-** to an ascii-string of 7 positions in milliseconds or minute-seconds or
-** hours-minutes, stored in strvalue (at least 8 positions).
+** to an ascii-string of 6 positions in milliseconds or minute-seconds or
+** hours-minutes, stored in strvalue (at least 7 positions).
 */
 #define	MAXMSEC		(count_t)100000
-#define	MAXSEC		(count_t)60000
+#define	MAXSEC		(count_t)6000
+#define	MAXMIN		(count_t)6000
 
 char *
 val2cpustr(count_t value, char *strvalue)
 {
 	if (value < MAXMSEC)
 	{
-		sprintf(strvalue, "%3lld.%02llds", value/1000, value%1000/10);
+		sprintf(strvalue, "%2lld.%02llds", value/1000, value%1000/10);
 	}
 	else
 	{
-		/*
-		** millisecs irrelevant; round to seconds
-		*/
-		value = (value + 500) / 1000;
+	        /*
+       	 	** millisecs irrelevant; round to seconds
+       	 	*/
+        	value = (value + 500) / 1000;
 
-		if (value < MAXSEC) 
-		{
-			sprintf(strvalue, "%3lldm%02llds", value/60, value%60);
+        	if (value < MAXSEC) 
+        	{
+               	 	sprintf(strvalue, "%2lldm%02llds", value/60, value%60);
 		}
 		else
 		{
@@ -274,13 +345,55 @@ val2cpustr(count_t value, char *strvalue)
 			*/
 			value = (value + 30) / 60;
 
-			sprintf(strvalue, "%3lldh%02lldm", value/60, value%60);
+			if (value < MAXMIN) 
+			{
+				sprintf(strvalue, "%2lldh%02lldm",
+							value/60, value%60);
+			}
+			else
+			{
+				/*
+				** minutes irrelevant; round to hours
+				*/
+				value = (value + 30) / 60;
+
+				sprintf(strvalue, "%2lldd%02lldh",
+						value/24, value%24);
+			}
 		}
 	}
 
 	return strvalue;
 }
 
+/*
+** Function val2Hzstr() converts a value (in MHz) 
+** to an ascii-string.
+** The result-string is placed in the area pointed to strvalue,
+** which should be able to contain at least 8 positions.
+*/
+char *
+val2Hzstr(count_t value, char *strvalue)
+{
+        if (value < 1000)
+        {
+                sprintf(strvalue, "%4lldMHz", value);
+        }
+        else
+        {
+                double fval=value/1000.0;      // fval is double in GHz
+                char prefix='G';
+
+                if (fval >= 1000.0)            // prepare for the future
+                {
+                        prefix='T';        
+                        fval /= 1000.0;
+                }
+                sprintf(strvalue, "%4.2f%cHz", fval, prefix);
+        }
+	return strvalue;
+}
+
 
 /*
 ** Function val2memstr() converts a value (number of bytes)
@@ -291,10 +404,12 @@ val2cpustr(count_t value, char *strvalue)
 #define	ONEKBYTE	1024
 #define	ONEMBYTE	1048576
 #define	ONEGBYTE	1073741824L
+#define	ONETBYTE	1099511627776LL
 
 #define	MAXBYTE		1024
 #define	MAXKBYTE	ONEKBYTE*99999L
 #define	MAXMBYTE	ONEMBYTE*999L
+#define	MAXGBYTE	ONEGBYTE*999LL
 
 char *
 val2memstr(count_t value, char *strvalue, int pformat, int avgval, int nsecs)
@@ -338,7 +453,10 @@ val2memstr(count_t value, char *strvalue, int pformat, int avgval, int nsecs)
 			if (verifyval <= MAXMBYTE)	/* mbytes ? */
 				aformat = MBFORMAT;
 			else
-				aformat = GBFORMAT;
+				if (verifyval <= MAXGBYTE)	/* mbytes ? */
+					aformat = GBFORMAT;
+				else
+					aformat = TBFORMAT;
 
 	/*
 	** check if this is also the preferred format
@@ -368,6 +486,11 @@ val2memstr(count_t value, char *strvalue, int pformat, int avgval, int nsecs)
 			basewidth-1, (double)((double)value/ONEGBYTE), suffix);
 		break;
 
+	   case	TBFORMAT:
+		sprintf(strvalue, "%*.1lfT%s",
+			basewidth-1, (double)((double)value/ONETBYTE), suffix);
+		break;
+
 	   default:
 		sprintf(strvalue, "!TILT!");
 	}
@@ -397,41 +520,46 @@ numeric(char *ns)
 
 /*
 ** Function getboot() returns the boot-time of this system 
-** (as number of seconds since 1-1-1970).
+** as number of jiffies since 1-1-1970.
 */
-time_t
+unsigned long long
 getboot(void)
 {
-	static time_t	boottime;
+	static unsigned long long	boottime;
+	unsigned long long		getbootlinux(long);
 
 	if (!boottime)		/* do this only once */
-	{
-#ifdef	linux
-		time_t		getbootlinux(long);
-
 		boottime = getbootlinux(hertz);
-#else
-		struct tms	tms;
-		time_t 		boottime_again;
-
-		/*
-		** beware that between time() and times() a one-second
-		** upgrade could have taken place in the kernel, so an
-		** extra check is issued; if we were around a
-		** second-upgrade, just do it again (we just passed
-		** the danger-zone)
-		*/
-		boottime 	= time(0) - (times(&tms) / hertz);
-		boottime_again	= time(0) - (times(&tms) / hertz);
-
-		if (boottime != boottime_again)
-			boottime = time(0) - (times(&tms) / hertz);
-#endif
-	}
 
 	return boottime;
 }
 
+/*
+** generic pointer verification after malloc
+*/
+void
+ptrverify(const void *ptr, const char *errormsg, ...)
+{
+	va_list args;
+
+        va_start(args, errormsg);
+
+	if (!ptr)
+	{
+		acctswoff();
+		netatop_signoff();
+
+		if (vis.show_end)
+			(vis.show_end)();
+
+        	va_list args;
+		fprintf(stderr, errormsg, args);
+        	va_end  (args);
+
+		exit(13);
+	}
+}
+
 /*
 ** signal catcher for cleanup before exit
 */
@@ -439,6 +567,33 @@ void
 cleanstop(exitcode)
 {
 	acctswoff();
+	netatop_signoff();
 	(vis.show_end)();
 	exit(exitcode);
 }
+
+/*
+** drop the root privileges that might be obtained via setuid-bit
+**
+** this action may only fail with errno EPERM (normal situation when
+** atop has not been started with setuid-root privs); when this
+** action fails with EAGAIN or ENOMEM, atop should not continue
+** without root privs being dropped...
+*/
+int
+droprootprivs(void)
+{
+	if (seteuid( getuid() ) == -1 && errno != EPERM)
+		return 0;	/* false */
+	else
+		return 1;	/* true  */
+}
+
+/*
+** regain the root privileges that might be dropped earlier
+*/
+void
+regainrootprivs(void)
+{
+	seteuid(0);
+}
